import logging
import time
from uuid import uuid4

from hazelcast import six
from hazelcast.config import _ReliableTopicConfig, TopicOverloadPolicy
from hazelcast.core import MemberInfo, MemberVersion
from hazelcast.errors import (
    OperationTimeoutError,
    IllegalArgumentError,
    HazelcastClientNotActiveError,
    ClientOfflineError,
    HazelcastInstanceNotActiveError,
    DistributedObjectDestroyedError,
    TopicOverloadError,
)
from hazelcast.future import ImmediateFuture, Future
from hazelcast.proxy.base import Proxy, TopicMessage
from hazelcast.proxy.ringbuffer import OVERFLOW_POLICY_FAIL, OVERFLOW_POLICY_OVERWRITE
from hazelcast.serialization.objects import ReliableTopicMessage
from hazelcast.util import check_not_none

_INITIAL_BACKOFF = 0.1
_MAX_BACKOFF = 2.0

_RINGBUFFER_PREFIX = "_hz_rb_"

_UNKNOWN_MEMBER_VERSION = MemberVersion(0, 0, 0)

_logger = logging.getLogger(__name__)


class ReliableMessageListener(object):
    """A message listener for :class:`ReliableTopic`.

    A message listener will not be called concurrently (provided that it's
    not registered twice). So there is no need to synchronize access to
    the state it reads or writes.

    If a regular function is registered on a reliable topic, the message
    listener works fine, but it can't do much more than listen to messages.

    This is an enhanced version of that to better integrate with the reliable
    topic.

    **Durable Subscription**

    The ReliableMessageListener allows you to control where you want to start
    processing a message when the listener is registered. This makes it
    possible to create a durable subscription by storing the sequence of the
    last message and using this as the sequence id to start from.

    **Error handling**

    The ReliableMessageListener also gives the ability to deal with errors
    using the :func:`is_terminal` method. If a plain function is used, then it
    won't terminate on errors and it will keep on running. But in some
    cases it is better to stop running.

    **Global order**

    The ReliableMessageListener will always get all events in order (global
    order). It will not get duplicates and there will only be gaps if it is
    too slow. For more information see :func:`is_loss_tolerant`.

    **Delivery guarantees**

    Because the ReliableMessageListener controls which item it wants to
    continue from upon restart, it is very easy to provide an at-least-once
    or at-most-once delivery guarantee. The :func:`store_sequence` is always
    called before a message is processed; so it can be persisted on some
    non-volatile storage. When the :func:`retrieve_initial_sequence` returns
    the stored sequence, then an at-least-once delivery is implemented since
    the same item is now being processed twice. To implement an at-most-once
    delivery guarantee, add 1 to the stored sequence when the
    :func:`retrieve_initial_sequence` is called.
    """

    def on_message(self, message):
        """
        Invoked when a message is received for the added reliable topic.

        One should not block in this callback. If blocking is necessary,
        consider delegating that task to an executor or a thread pool.

        Args:
            message (hazelcast.proxy.base.TopicMessage): The message that
                is received for the topic
        """
        raise NotImplementedError("on_message")

    def retrieve_initial_sequence(self):
        """
        Retrieves the initial sequence from which this ReliableMessageListener
        should start.

        Return ``-1`` if there is no initial sequence and you want to start
        from the next published message.

        If you intend to create a durable subscriber so you continue from where
        you stopped the previous time, load the previous sequence and add ``1``.
        If you don't add one, then you will be receiving the same message twice.

        Returns:
            int: The initial sequence.
        """
        raise NotImplementedError("retrieve_initial_sequence")

    def store_sequence(self, sequence):
        """
        Informs the ReliableMessageListener that it should store the sequence.
        This method is called before the message is processed. Can be used to
        make a durable subscription.

        Args:
            sequence (int): The sequence.
        """
        raise NotImplementedError("store_sequence")

    def is_loss_tolerant(self):
        """
        Checks if this ReliableMessageListener is able to deal with message loss.
        Even though the reliable topic promises to be reliable, it can be that a
        ReliableMessageListener is too slow. Eventually the message won't be
        available anymore.

        If the ReliableMessageListener is not loss tolerant and the topic detects
        that there are missing messages, it will terminate the
        ReliableMessageListener.

        Returns:
            bool: ``True`` if the ReliableMessageListener is tolerant towards
            losing messages.
        """
        raise NotImplementedError("is_loss_tolerant")

    def is_terminal(self, error):
        """
        Checks if the ReliableMessageListener should be terminated based on an
        error raised while calling :func:`on_message`.

        Args:
            error (Exception): The error raised while calling
                :func:`on_message`

        Returns:
            bool: ``True`` if the ReliableMessageListener should terminate itself,
            ``False`` if it should keep on running.
        """
        raise NotImplementedError("is_terminal")


class _MessageRunner(object):
    def __init__(
        self,
        registration_id,
        listener,
        ringbuffer,
        topic_name,
        read_batch_size,
        to_object,
        runners,
    ):
        self._registration_id = registration_id
        self._listener = listener
        self._ringbuffer = ringbuffer
        self._topic_name = topic_name
        self._read_batch_size = read_batch_size
        self._to_object = to_object
        self._runners = runners
        self._sequence = listener.retrieve_initial_sequence()
        self._cancelled = False

    def start(self):
        """Starts the message runner by checking the given sequence.

        If the user provided a initial sequence via listener, we will
        use it as it is. If not, we will ask server to get the tail
        sequence and use it.

        Returns:
            hazelcast.future.Future[None]:
        """
        if self._sequence != -1:
            # User provided a sequence to start from
            return ImmediateFuture(None)

        def continuation(future):
            sequence = future.result()
            self._sequence = sequence + 1

        # We are going to listen to next publication.
        # We don't care about what already has been published.
        return self._ringbuffer.tail_sequence().continue_with(continuation)

    def next_batch(self):
        """Tries to read the next batch from the ringbuffer
        and call the listener on items when it is done.
        """
        if self._cancelled:
            return

        self._ringbuffer.read_many(self._sequence, 1, self._read_batch_size).add_done_callback(
            self._handle_next_batch
        )

    def cancel(self):
        """Sets the cancelled flag and removes
        the runner registration.
        """
        self._cancelled = True
        self._runners.pop(self._registration_id, None)

    def _handle_next_batch(self, future):
        """Handles the result of the read_many request from
        the ringbuffer.

        Args:
            future (hazelcast.future.Future):
        """
        if self._cancelled:
            return

        try:
            result = future.result()

            # Check if there are any messages lost since the last read
            # and whether or not the listener can tolerate that.
            lost_count = (result.next_sequence_to_read_from - result.read_count) - self._sequence
            if lost_count != 0 and not self._is_loss_tolerable(lost_count):
                self.cancel()
                return

            # Call the listener for each item read.
            for i in range(result.size):
                try:
                    message = result[i]
                    self._listener.store_sequence(result.get_sequence(i))

                    member = None
                    if message.publisher_address:
                        member = MemberInfo(
                            message.publisher_address, None, None, False, _UNKNOWN_MEMBER_VERSION
                        )

                    topic_message = TopicMessage(
                        self._topic_name,
                        message.payload,
                        message.publish_time,
                        member,
                        self._to_object,
                    )
                    self._listener.on_message(topic_message)
                except Exception as e:
                    if self._terminate(e):
                        self.cancel()
                        return

            self._sequence = result.next_sequence_to_read_from
            self.next_batch()
        except Exception as e:
            # read_many request failed.
            if not self._handle_internal_error(e):
                self.cancel()

    def _is_loss_tolerable(self, loss_count):
        """Called when message loss is detected.

        Checks if the listener is able to tolerate the loss.

        Args:
            loss_count (int): Number of lost messages.

        Returns:
            bool: ``True`` if the listener may continue reading.
        """
        if self._listener.is_loss_tolerant():
            _logger.debug(
                "MessageListener %s on topic %s lost %s messages.",
                self._listener,
                self._topic_name,
                loss_count,
            )
            return True

        _logger.warning(
            "Terminating MessageListener %s on topic %s. "
            "Reason: The listener was too slow or the retention period of the message has been violated. "
            "%s messages lost.",
            self._listener,
            self._topic_name,
            loss_count,
        )
        return False

    def _terminate(self, error):
        """Checks if we should terminate the listener
        based on the error we received while calling the
        on_message for this message.

        If the listener says that it should be terminated
        for this error or it raises some error while
        we were trying to call is_terminal, the listener
        will be terminated. Otherwise, a log will be
        printed and listener will continue.

        Args:
            error (Exception): Error we received while
                calling the listener.

        Returns:
            bool: Should terminate the listener or not.
        """
        if self._cancelled:
            return True

        try:
            terminate = self._listener.is_terminal(error)
            if terminate:
                _logger.warning(
                    "Terminating MessageListener %s on topic %s. Reason: Unhandled exception.",
                    self._listener,
                    self._topic_name,
                    exc_info=error,
                )
            else:
                _logger.debug(
                    "MessageListener %s on topic %s ran into an error.",
                    self._listener,
                    self._topic_name,
                    exc_info=error,
                )
            return terminate
        except Exception as e:
            _logger.warning(
                "Terminating MessageListener %s on topic %s. "
                "Reason: Unhandled exception while calling is_terminal method",
                self._listener,
                self._topic_name,
                exc_info=e,
            )
            return True

    def _handle_internal_error(self, error):
        """Called when the read_many request is failed.

        Based on the error we receive, we will act differently.

        If we can tolerate the error, we will call next_batch
        here. The reasoning behind is that, on some cases, we
        do not immediately call next_batch, but make a request
        to the server, and based on that, call next_batch.

        Args:
            error (Exception): The error we received.

        Returns:
            bool: ``True`` if the error is handled internally.
            ``False`` otherwise. When, ``False`` is returned,
            listener should be cancelled.
        """
        if isinstance(error, HazelcastClientNotActiveError):
            return self._handle_client_not_active_error()
        elif isinstance(error, ClientOfflineError):
            return self._handle_client_offline_error()
        elif isinstance(error, OperationTimeoutError):
            return self._handle_timeout_error()
        elif isinstance(error, IllegalArgumentError):
            return self._handle_illegal_argument_error(error)
        elif isinstance(error, HazelcastInstanceNotActiveError):
            return self._handle_instance_not_active_error()
        elif isinstance(error, DistributedObjectDestroyedError):
            return self._handle_distributed_object_destroyed_error()
        else:
            return self._handle_generic_error(error)

    def _handle_generic_error(self, error):
        # Received an error we do not expect.
        _logger.warning(
            "Terminating MessageListener %s on topic %s. Reason: Unhandled exception.",
            self._listener,
            self._topic_name,
            exc_info=error,
        )
        return False

    def _handle_distributed_object_destroyed_error(self):
        # Underlying ringbuffer is destroyed. It should only
        # happen when the user destroys the reliable topic
        # associated with it.
        _logger.debug(
            "Terminating MessageListener %s on topic %s. Reason: Topic is destroyed.",
            self._listener,
            self._topic_name,
        )
        return False

    def _handle_instance_not_active_error(self):
        # This error should be received from the server.
        # We do not throw it anywhere on the client.
        _logger.debug(
            "Terminating MessageListener %s on topic %s. Reason: Server is shutting down.",
            self._listener,
            self._topic_name,
        )
        return False

    def _handle_client_offline_error(self):
        # Client is reconnecting to cluster.
        _logger.debug(
            "MessageListener %s on topic %s got error. "
            "Continuing from the last known sequence %s.",
            self._listener,
            self._topic_name,
            self._sequence,
        )
        self.next_batch()
        return True

    def _handle_client_not_active_error(self):
        # Client#shutdown is called.
        _logger.debug(
            "Terminating MessageListener %s on topic %s. Reason: Client is shutting down.",
            self._listener,
            self._topic_name,
        )
        return False

    def _handle_timeout_error(self):
        # read_many invocation to the server timed out.
        _logger.debug(
            "MessageListener %s on topic %s timed out. "
            "Continuing from the last known sequence %s.",
            self._listener,
            self._topic_name,
            self._sequence,
        )
        self.next_batch()
        return True

    def _handle_illegal_argument_error(self, error):
        # Server sends this when it detects data loss
        # on the underlying ringbuffer.
        if self._listener.is_loss_tolerant():
            # Listener can tolerate message loss. Try
            # to continue reading after getting head
            # sequence, and try to read from there.
            def on_response(future):
                try:
                    head_sequence = future.result()
                    _logger.debug(
                        "MessageListener %s on topic %s requested a too large sequence. "
                        "Jumping from old sequence %s to sequence %s.",
                        self._listener,
                        self._topic_name,
                        self._sequence,
                        head_sequence,
                        exc_info=error,
                    )
                    self._sequence = head_sequence
                    # We call next_batch only after getting the new head
                    # sequence and updating the our state with it.
                    self.next_batch()
                except Exception as e:
                    _logger.warning(
                        "Terminating MessageListener %s on topic %s. "
                        "Reason: After the ring buffer data related "
                        "to reliable topic is lost, client tried to get the "
                        "current head sequence to continue since the listener"
                        "is loss tolerant, but that request failed.",
                        self._listener,
                        self._topic_name,
                        exc_info=e,
                    )
                    # We said that we can handle that error so the listener
                    # is not cancelled. But, we could not continue since
                    # our request to the server is failed. We should cancel
                    # the listener.
                    self.cancel()

            self._ringbuffer.head_sequence().add_done_callback(on_response)
            return True

        _logger.warning(
            "Terminating MessageListener %s on topic %s. "
            "Reason: Underlying ring buffer data related to reliable topic is lost.",
            self._listener,
            self._topic_name,
        )

        return False


class _ReliableMessageListenerAdapter(ReliableMessageListener):
    """Used when the user provided a function as the listener.

    That means user does not care about the other properties
    of the listener. They just want to listen messages. Fill
    the methods with expected defaults.
    """

    def __init__(self, on_message_fn):
        self._on_message_fn = on_message_fn

    def on_message(self, message):
        self._on_message_fn(message)

    def retrieve_initial_sequence(self):
        # -1 indicates start from next message.
        return -1

    def store_sequence(self, sequence):
        # no-op
        pass

    def is_loss_tolerant(self):
        # terminate the listener on message loss
        return False

    def is_terminal(self, error):
        # do not terminate the listener or errors
        return False


def _no_op_continuation(future):
    # Used when we just care about whether
    # the ringbuffer request is failed. We just
    # check the result and return nothing.
    future.result()


class ReliableTopic(Proxy):
    """Hazelcast provides distribution mechanism for publishing messages that
    are delivered to multiple subscribers, which is also known as a
    publish/subscribe (pub/sub) messaging model. Publish and subscriptions are
    cluster-wide. When a member subscribes for a topic, it is actually
    registering for messages published by any member in the cluster, including
    the new members joined after you added the listener.

    Messages are ordered, meaning that listeners(subscribers) will process the
    messages in the order they are actually published.

    Hazelcast's Reliable Topic uses the same Topic interface as a regular topic.
    The main difference is that Reliable Topic is backed up by the Ringbuffer
    data structure, a replicated but not partitioned data structure that stores
    its data in a ring-like structure.
    """

    def __init__(self, service_name, name, context):
        super(ReliableTopic, self).__init__(service_name, name, context)

        config = context.config.reliable_topics.get(name, None)
        if config is None:
            config = _ReliableTopicConfig()

        self._config = config
        self._context = context
        self._ringbuffer = context.client.get_ringbuffer(_RINGBUFFER_PREFIX + name)
        self._runners = {}

    def publish(self, message):
        """Publishes the message to all subscribers of this topic.

        Args:
            message: The message.

        Returns:
            hazelcast.future.Future[None]:
        """
        check_not_none(message, "Message cannot be None")

        payload = self._to_data(message)
        topic_message = ReliableTopicMessage(time.time(), None, payload)

        overload_policy = self._config.overload_policy
        if overload_policy == TopicOverloadPolicy.BLOCK:
            return self._add_with_backoff(topic_message)
        elif overload_policy == TopicOverloadPolicy.ERROR:
            return self._add_or_fail(topic_message)
        elif overload_policy == TopicOverloadPolicy.DISCARD_OLDEST:
            return self._add_or_overwrite(topic_message)
        elif overload_policy == TopicOverloadPolicy.DISCARD_NEWEST:
            return self._add_or_discard(topic_message)

    def publish_all(self, messages):
        """Publishes all messages to all subscribers of this topic.

        Args:
            messages (list): Messages to publish.

        Returns:
            hazelcast.future.Future[None]:
        """
        check_not_none(messages, "Messages cannot be None")

        topic_messages = []

        for message in messages:
            check_not_none(message, "Message cannot be None")
            payload = self._to_data(message)
            topic_messages.append(ReliableTopicMessage(time.time(), None, payload))

        overload_policy = self._config.overload_policy
        if overload_policy == TopicOverloadPolicy.BLOCK:
            return self._add_messages_with_backoff(topic_messages)
        elif overload_policy == TopicOverloadPolicy.ERROR:
            return self._add_messages_or_fail(topic_messages)
        elif overload_policy == TopicOverloadPolicy.DISCARD_OLDEST:
            return self._add_messages_or_overwrite(topic_messages)
        elif overload_policy == TopicOverloadPolicy.DISCARD_NEWEST:
            return self._add_messages_or_discard(topic_messages)

    def add_listener(self, listener):
        """Subscribes to this reliable topic.

        It can be either a simple function or an instance of an
        :class:`ReliableMessageListener`. When a function is passed, a
        :class:`ReliableMessageListener` is created out of that with
        sensible default values.

        When a message is published, the,
        :func:`ReliableMessageListener.on_message` method of the given
        listener (or the function passed) is called.

        More than one message listener can be added on one instance.

        Args:
            listener (function or ReliableMessageListener): Listener to add.

        Returns:
            hazelcast.future.Future[str]: The registration id.
        """
        check_not_none(listener, "None listener is not allowed")

        registration_id = str(uuid4())
        reliable_message_listener = self._to_reliable_message_listener(listener)

        runner = _MessageRunner(
            registration_id,
            reliable_message_listener,
            self._ringbuffer,
            self.name,
            self._config.read_batch_size,
            self._to_object,
            self._runners,
        )

        def continuation(future):
            future.result()
            # If the runner started successfully, register it.
            self._runners[registration_id] = runner
            runner.next_batch()
            return registration_id

        return runner.start().continue_with(continuation)

    def remove_listener(self, registration_id):
        """Stops receiving messages for the given message listener.

        If the given listener already removed, this method does nothing.

        Args:
            registration_id (str): ID of listener registration.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if registration is
            removed, ``False`` otherwise.
        """
        check_not_none(registration_id, "Registration id cannot be None")
        runner = self._runners.get(registration_id, None)
        if not runner:
            return ImmediateFuture(False)

        runner.cancel()
        return ImmediateFuture(True)

    def destroy(self):
        """
        Destroys underlying Proxy and RingBuffer instances.
        """

        for runner in list(six.itervalues(self._runners)):
            runner.cancel()

        self._runners.clear()

        super(ReliableTopic, self).destroy()
        return self._ringbuffer.destroy()

    def _add_or_fail(self, message):
        def continuation(future):
            sequence_id = future.result()
            if sequence_id == -1:
                raise TopicOverloadError(
                    "Failed to publish message %s on topic %s." % (message, self.name)
                )

        return self._ringbuffer.add(message, OVERFLOW_POLICY_FAIL).continue_with(continuation)

    def _add_messages_or_fail(self, messages):
        def continuation(future):
            sequence_id = future.result()
            if sequence_id == -1:
                raise TopicOverloadError("Failed to publish messages on topic %s." % self.name)

        return self._ringbuffer.add_all(messages, OVERFLOW_POLICY_FAIL).continue_with(continuation)

    def _add_or_overwrite(self, message):
        return self._ringbuffer.add(message, OVERFLOW_POLICY_OVERWRITE).continue_with(
            _no_op_continuation
        )

    def _add_messages_or_overwrite(self, messages):
        return self._ringbuffer.add_all(messages, OVERFLOW_POLICY_OVERWRITE).continue_with(
            _no_op_continuation
        )

    def _add_or_discard(self, message):
        return self._ringbuffer.add(message, OVERFLOW_POLICY_FAIL).continue_with(
            _no_op_continuation
        )

    def _add_messages_or_discard(self, messages):
        return self._ringbuffer.add_all(messages, OVERFLOW_POLICY_FAIL).continue_with(
            _no_op_continuation
        )

    def _add_with_backoff(self, message):
        future = Future()
        self._try_adding_with_backoff(message, _INITIAL_BACKOFF, future)
        return future

    def _add_messages_with_backoff(self, messages):
        future = Future()
        self._try_adding_messages_with_backoff(messages, _INITIAL_BACKOFF, future)
        return future

    def _try_adding_with_backoff(self, message, backoff, future):
        def callback(add_future):
            try:
                sequence_id = add_future.result()
                if sequence_id != -1:
                    future.set_result(None)
                    return

                self._context.reactor.add_timer(
                    backoff,
                    lambda: self._try_adding_with_backoff(
                        message, min(_MAX_BACKOFF, 2 * backoff), future
                    ),
                )
            except Exception as e:
                future.set_result(e)

        self._ringbuffer.add(message, OVERFLOW_POLICY_FAIL).add_done_callback(callback)

    def _try_adding_messages_with_backoff(self, messages, backoff, future):
        def callback(add_future):
            try:
                sequence_id = add_future.result()
                if sequence_id != -1:
                    future.set_result(None)
                    return

                self._context.reactor.add_timer(
                    backoff,
                    lambda: self._try_adding_messages_with_backoff(
                        messages, min(_MAX_BACKOFF, 2 * backoff), future
                    ),
                )
            except Exception as e:
                future.set_result(e)

        self._ringbuffer.add_all(messages, OVERFLOW_POLICY_FAIL).add_done_callback(callback)

    @staticmethod
    def _to_reliable_message_listener(listener):
        if isinstance(listener, ReliableMessageListener):
            return listener

        if not callable(listener):
            raise TypeError("Listener must be a callable")

        return _ReliableMessageListenerAdapter(listener)



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import DataCodec

# hex: 0x021400
_REQUEST_MESSAGE_TYPE = 136192
# hex: 0x021401
_RESPONSE_MESSAGE_TYPE = 136193

_REQUEST_REFERENCE_ID_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_REFERENCE_ID_OFFSET + LONG_SIZE_IN_BYTES


def encode_request(name, key, reference_id):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_REFERENCE_ID_OFFSET, reference_id)
    StringCodec.encode(buf, name)
    DataCodec.encode(buf, key, True)
    return OutboundMessage(buf, True)



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer
from hazelcast.protocol.codec.custom.raft_group_id_codec import RaftGroupIdCodec
from hazelcast.protocol.builtin import StringCodec

# hex: 0x0B0300
_REQUEST_MESSAGE_TYPE = 721664
# hex: 0x0B0301
_RESPONSE_MESSAGE_TYPE = 721665

_REQUEST_INVOCATION_UID_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_EXPECTED_ROUND_OFFSET = _REQUEST_INVOCATION_UID_OFFSET + UUID_SIZE_IN_BYTES
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_EXPECTED_ROUND_OFFSET + INT_SIZE_IN_BYTES


def encode_request(group_id, name, invocation_uid, expected_round):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_uuid(buf, _REQUEST_INVOCATION_UID_OFFSET, invocation_uid)
    FixSizedTypesCodec.encode_int(buf, _REQUEST_EXPECTED_ROUND_OFFSET, expected_round)
    RaftGroupIdCodec.encode(buf, group_id)
    StringCodec.encode(buf, name, True)
    return OutboundMessage(buf, True)



import asyncore
import errno
import io
import logging
import os
import select
import socket
import ssl
import sys
import threading
import time

from collections import deque
from functools import total_ordering
from heapq import heappush, heappop

from hazelcast.config import SSLProtocol
from hazelcast.connection import Connection
from hazelcast.core import Address
from hazelcast.errors import HazelcastError
from hazelcast.future import Future

try:
    import fcntl
except ImportError:
    fcntl = None

try:
    from _thread import get_ident
except ImportError:
    # Python2
    from thread import get_ident

_logger = logging.getLogger(__name__)

# We should retry receiving/sending the message in case of these errors
# EAGAIN: Resource temporarily unavailable
# EWOULDBLOCK: The read/write would block
# EDEADLK: Was added before, retrying it just to make sure that
#   client behaves the same on some edge cases.
# SSL_ERROR_WANT_READ/WRITE: The socket could not satisfy the
#   needs of the SSL_read/write. During the negotiation process
#   SSL_read/write may also want to write/read data, hence may also
#   raise SSL_ERROR_WANT_WRITE/READ.
_RETRYABLE_ERROR_CODES = (
    errno.EAGAIN,
    errno.EWOULDBLOCK,
    errno.EDEADLK,
    ssl.SSL_ERROR_WANT_WRITE,
    ssl.SSL_ERROR_WANT_READ,
)


def _set_nonblocking(fd):
    if not fcntl:
        return

    flags = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)


class _SocketAdapter(object):
    def __init__(self, fd):
        self._fd = fd

    def fileno(self):
        return self._fd

    def close(self):
        os.close(self._fd)

    def getsockopt(self, level, optname, buflen=None):
        if level == socket.SOL_SOCKET and optname == socket.SO_ERROR and not buflen:
            return 0
        raise NotImplementedError("Only asyncore specific behaviour is implemented.")


class _AbstractWaker(asyncore.dispatcher):
    def __init__(self, map):
        asyncore.dispatcher.__init__(self, map=map)
        self.awake = False

    def writable(self):
        return False

    def wake(self):
        raise NotImplementedError("wake")


class _PipedWaker(_AbstractWaker):
    def __init__(self, map):
        _AbstractWaker.__init__(self, map)
        self._read_fd, self._write_fd = os.pipe()
        self.set_socket(_SocketAdapter(self._read_fd))
        _set_nonblocking(self._read_fd)
        _set_nonblocking(self._write_fd)

    def wake(self):
        if not self.awake:
            self.awake = True
            try:
                os.write(self._write_fd, b"x")
            except (IOError, ValueError):
                pass

    def handle_read(self):
        self.awake = False
        try:
            while len(os.read(self._read_fd, 4096)) == 4096:
                pass
        except (IOError, OSError):
            pass

    def close(self):
        _AbstractWaker.close(self)  # Will close the reader
        os.close(self._write_fd)


class _SocketedWaker(_AbstractWaker):
    def __init__(self, map):
        _AbstractWaker.__init__(self, map)
        self._writer = socket.socket()
        self._writer.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        a = socket.socket()
        a.bind(("127.0.0.1", 0))
        a.listen(1)
        addr = a.getsockname()

        try:
            self._writer.connect(addr)
            self._reader, _ = a.accept()
        finally:
            a.close()

        self.set_socket(self._reader)
        self._writer.settimeout(0)
        self._reader.settimeout(0)

    def wake(self):
        if not self.awake:
            self.awake = True
            try:
                self._writer.send(b"x")
            except (IOError, socket.error, ValueError):
                pass

    def handle_read(self):
        self.awake = False
        try:
            while len(self._reader.recv(4096)) == 4096:
                pass
        except (IOError, socket.error):
            pass

    def close(self):
        _AbstractWaker.close(self)  # Will close the reader
        self._writer.close()


class _AbstractLoop(object):
    def __init__(self, map):
        self._map = map
        self._timers = []  # Accessed only from the reactor thread
        self._new_timers = deque()  # Popped only from the reactor thread
        self._is_live = False
        self._thread = None
        self._ident = -1

    def start(self):
        self._is_live = True
        self._thread = threading.Thread(target=self._loop, name="hazelcast-reactor")
        self._thread.daemon = True
        self._thread.start()
        self._ident = self._thread.ident

    def _loop(self):
        _logger.debug("Starting Reactor Thread")
        Future._threading_locals.is_reactor_thread = True
        while self._is_live:
            try:
                self.run_loop()
                self._check_timers()
            except select.error:
                # TODO: parse error type to catch only error "9"
                _logger.warning("Connection closed by server")
                pass
            except:
                _logger.exception("Error in Reactor Thread")
                # TODO: shutdown client
                return
        _logger.debug("Reactor Thread exited")
        self._cleanup_all_timers()

    def add_timer(self, delay, callback):
        timer = Timer(delay + time.time(), callback)
        self._new_timers.append((timer.end, timer))
        return timer

    def _check_timers(self):
        timers = self._timers

        if self._new_timers:
            new_timers = self._new_timers
            while new_timers:
                # There is no need to check for exception here,
                # reactor thread is the only one popping from
                # the deque. So, if the we pass the size check
                # above, there should be at least one element
                heappush(timers, new_timers.popleft())

        if timers:
            now = time.time()
            while timers:
                timer = timers[0][1]
                if timer.check_timer(now):
                    heappop(timers)
                else:
                    # Timer in the root of the min heap is not expired.
                    # Therefore, there should be no expired
                    # timers in the heap.
                    return

    def _cleanup_all_timers(self):
        timers = self._timers
        new_timers = self._new_timers

        while timers:
            _, timer = timers.pop()
            timer.timer_ended_cb()

        # Although it is not the case with the current code base,
        # the timers ended above may add new timers. So, the order
        # is important.
        while new_timers:
            _, timer = new_timers.popleft()
            timer.timer_ended_cb()

    def check_loop(self):
        raise NotImplementedError("check_loop")

    def run_loop(self):
        raise NotImplementedError("run_loop")

    def wake_loop(self):
        raise NotImplementedError("wake_loop")

    def shutdown(self):
        raise NotImplementedError("shutdown")


class _WakeableLoop(_AbstractLoop):
    _waker_class = _PipedWaker if os.name != "nt" else _SocketedWaker

    def __init__(self, map):
        _AbstractLoop.__init__(self, map)
        self.waker = self._waker_class(map)

    def check_loop(self):
        assert not self.waker.awake
        self.wake_loop()
        assert self.waker.awake
        self.run_loop()
        assert not self.waker.awake

    def run_loop(self):
        asyncore.loop(timeout=0.01, use_poll=True, map=self._map, count=1)

    def wake_loop(self):
        if self._ident != get_ident():
            self.waker.wake()

    def shutdown(self):
        if not self._is_live:
            return

        self._is_live = False

        if self._ident != get_ident():
            self._thread.join()

        for connection in list(self._map.values()):
            if connection is self.waker:
                continue

            try:
                connection.close(None, HazelcastError("Client is shutting down"))
            except OSError as connection:
                if connection.args[0] == socket.EBADF:
                    pass
                else:
                    raise

        self.waker.close()
        self._map.clear()


class _BasicLoop(_AbstractLoop):
    def check_loop(self):
        pass

    def run_loop(self):
        asyncore.loop(timeout=0.001, use_poll=True, map=self._map, count=1)

    def wake_loop(self):
        pass

    def shutdown(self):
        if not self._is_live:
            return

        self._is_live = False

        if self._ident != get_ident():
            self._thread.join()

        for connection in list(self._map.values()):
            try:
                connection.close(None, HazelcastError("Client is shutting down"))
            except OSError as connection:
                if connection.args[0] == socket.EBADF:
                    pass
                else:
                    raise

        self._map.clear()


class AsyncoreReactor(object):
    def __init__(self):
        self.map = {}
        loop = None
        try:
            loop = _WakeableLoop(self.map)
            loop.check_loop()
        except:
            _logger.exception(
                "Failed to initialize the wakeable loop. "
                "Using the basic loop instead. "
                "When used in the blocking mode, client"
                "may have sub-optimal performance."
            )
            if loop:
                loop.shutdown()
            loop = _BasicLoop(self.map)
        self._loop = loop

    def start(self):
        self._loop.start()

    def add_timer(self, delay, callback):
        return self._loop.add_timer(delay, callback)

    def wake_loop(self):
        self._loop.wake_loop()

    def shutdown(self):
        self._loop.shutdown()

    def connection_factory(
        self, connection_manager, connection_id, address, network_config, message_callback
    ):
        return AsyncoreConnection(
            self, connection_manager, connection_id, address, network_config, message_callback
        )


_BUFFER_SIZE = 128000


class AsyncoreConnection(Connection, asyncore.dispatcher):
    sent_protocol_bytes = False
    receive_buffer_size = _BUFFER_SIZE
    send_buffer_size = _BUFFER_SIZE
    _close_timer = None

    def __init__(
        self, reactor, connection_manager, connection_id, address, config, message_callback
    ):
        asyncore.dispatcher.__init__(self, map=reactor.map)
        Connection.__init__(self, connection_manager, connection_id, message_callback)

        self._reactor = reactor
        self.connected_address = address
        self._write_queue = deque()
        self._write_buf = io.BytesIO()

        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        # set the socket timeout to 0 explicitly
        self.socket.settimeout(0)
        self._set_socket_options(config)
        if config.ssl_enabled:
            self._wrap_as_ssl_socket(config)

        try:
            self.connect((address.host, address.port))
        except socket.error as e:
            # If the connection attempt failed
            # immediately, remove the connection from
            # the dispatchers map and clean resources.
            self._inner_close()
            raise e

        timeout = config.connection_timeout
        if timeout > 0:
            self._close_timer = reactor.add_timer(timeout, self._close_timer_cb)

        self.local_address = Address(*self.socket.getsockname())
        self._write_queue.append(b"CP2")

    def handle_connect(self):
        if self._close_timer:
            self._close_timer.cancel()

        self.start_time = time.time()
        _logger.debug("Connected to %s", self.connected_address)

    def handle_read(self):
        reader = self._reader
        receive_buffer_size = self.receive_buffer_size
        try:
            while True:
                data = self.recv(receive_buffer_size)
                reader.read(data)
                self.last_read_time = time.time()
                if len(data) < receive_buffer_size:
                    break
        except socket.error as err:
            if err.args[0] not in _RETRYABLE_ERROR_CODES:
                # Other error codes are fatal, should close the connection
                self.close(None, err)

        if reader.length:
            reader.process()

    def handle_write(self):
        write_queue = self._write_queue
        send_buffer_size = self.send_buffer_size
        write_batch = []
        total_length = 0

        while write_queue:
            message_bytes = write_queue.popleft()
            write_batch.append(message_bytes)
            total_length += len(message_bytes)

            if total_length >= send_buffer_size:
                break

        # We enter this only if len(write_queue) > 0.
        # So, len(write_batch) cannot be 0.
        if len(write_batch) == 1:
            bytes_ = write_batch[0]
        else:
            buf = self._write_buf
            buf.seek(0)
            for message_bytes in write_batch:
                buf.write(message_bytes)

            bytes_ = buf.getvalue()
            buf.truncate(0)

        try:
            sent = self.send(bytes_)
        except socket.error as err:
            if err.args[0] in _RETRYABLE_ERROR_CODES:
                # Couldn't write the bytes but we should
                # retry it.
                self._write_queue.appendleft(bytes_)
            else:
                # Other error codes are fatal, should close the connection
                self.close(None, err)
        else:
            # No exception is thrown during the send
            self.last_write_time = time.time()
            self.sent_protocol_bytes = True

            if sent < len(bytes_):
                write_queue.appendleft(bytes_[sent:])

    def handle_close(self):
        _logger.warning("Connection closed by server")
        self.close(None, IOError("Connection closed by server"))

    def handle_error(self):
        # We handle retryable error codes inside the
        # handle_read/write. Anything else should be fatal.
        error = sys.exc_info()[1]
        _logger.debug("Received error", exc_info=True)
        self.close(str(error), None)

    def readable(self):
        return self.live and self.sent_protocol_bytes

    def _write(self, buf):
        self._write_queue.append(buf)
        self._reactor.wake_loop()

    def writable(self):
        return len(self._write_queue) > 0

    def _inner_close(self):
        if self._close_timer:
            # It might be the case that connection
            # is closed before the timer. If we are
            # closing via the timer, this call has
            # no effects.
            self._close_timer.cancel()

        asyncore.dispatcher.close(self)
        self._write_buf.close()

    def _close_timer_cb(self):
        if not self.connected:
            self.close(None, IOError("Connection timed out"))

    def _set_socket_options(self, config):
        # set tcp no delay
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        # set socket buffer
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, _BUFFER_SIZE)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, _BUFFER_SIZE)

        for level, option_name, value in config.socket_options:
            if option_name is socket.SO_RCVBUF:
                self.receive_buffer_size = value
            elif option_name is socket.SO_SNDBUF:
                self.send_buffer_size = value

            self.socket.setsockopt(level, option_name, value)

    def _wrap_as_ssl_socket(self, config):
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)

        protocol = config.ssl_protocol

        # Use only the configured protocol
        try:
            if protocol != SSLProtocol.SSLv2:
                ssl_context.options |= ssl.OP_NO_SSLv2
            if protocol != SSLProtocol.SSLv3:
                ssl_context.options |= ssl.OP_NO_SSLv3
            if protocol != SSLProtocol.TLSv1:
                ssl_context.options |= ssl.OP_NO_TLSv1
            if protocol != SSLProtocol.TLSv1_1:
                ssl_context.options |= ssl.OP_NO_TLSv1_1
            if protocol != SSLProtocol.TLSv1_2:
                ssl_context.options |= ssl.OP_NO_TLSv1_2
            if protocol != SSLProtocol.TLSv1_3:
                ssl_context.options |= ssl.OP_NO_TLSv1_3
        except AttributeError:
            pass

        ssl_context.verify_mode = ssl.CERT_REQUIRED

        if config.ssl_cafile:
            ssl_context.load_verify_locations(config.ssl_cafile)
        else:
            ssl_context.load_default_certs()

        if config.ssl_certfile:
            ssl_context.load_cert_chain(
                config.ssl_certfile, config.ssl_keyfile, config.ssl_password
            )

        if config.ssl_ciphers:
            ssl_context.set_ciphers(config.ssl_ciphers)

        self.socket = ssl_context.wrap_socket(self.socket)

    def __repr__(self):
        return "Connection(id=%s, live=%s, remote_address=%s)" % (
            self._id,
            self.live,
            self.remote_address,
        )

    def __str__(self):
        return self.__repr__()


@total_ordering
class Timer(object):
    __slots__ = ("end", "timer_ended_cb", "canceled")

    def __init__(self, end, timer_ended_cb):
        self.end = end
        self.timer_ended_cb = timer_ended_cb
        self.canceled = False

    def __eq__(self, other):
        return self.end == other.end

    def __ne__(self, other):
        return self.end != other.end

    def __lt__(self, other):
        return self.end < other.end

    def cancel(self):
        self.canceled = True

    def check_timer(self, now):
        if self.canceled:
            return True

        if now >= self.end:
            self.timer_ended_cb()
            return True

        return False



import re
import types

from hazelcast import six
from hazelcast.errors import InvalidConfigurationError
from hazelcast.serialization.api import StreamSerializer, IdentifiedDataSerializable, Portable
from hazelcast.serialization.portable.classdef import ClassDefinition
from hazelcast.security import TokenProvider
from hazelcast.util import (
    check_not_none,
    number_types,
    LoadBalancer,
    none_type,
    try_to_get_enum_value,
)


class IntType(object):
    """Integer type options that can be used by serialization service."""

    VAR = 0
    """
    Integer types will be serialized as 8, 16, 32, 64 bit integers
    or as Java BigInteger according to their value. This option may
    cause problems when the Python client is used in conjunction with
    statically typed language clients such as Java or .NET.
    """

    BYTE = 1
    """
    Integer types will be serialized as a 8 bit integer(as Java byte)
    """

    SHORT = 2
    """
    Integer types will be serialized as a 16 bit integer(as Java short)
    """

    INT = 3
    """
    Integer types will be serialized as a 32 bit integer(as Java int)
    """

    LONG = 4
    """
    Integer types will be serialized as a 64 bit integer(as Java long)
    """

    BIG_INT = 5
    """
    Integer types will be serialized as Java BigInteger. This option can
    handle integer types which are less than -2^63 or greater than or
    equal to 2^63. However, when this option is set, serializing/de-serializing
    integer types is costly.
    """


class EvictionPolicy(object):
    """Near Cache eviction policy options."""

    NONE = 0
    """
    No eviction.
    """

    LRU = 1
    """
    Least Recently Used items will be evicted.
    """

    LFU = 2
    """
    Least frequently Used items will be evicted.
    """

    RANDOM = 3
    """
    Items will be evicted randomly.
    """


class InMemoryFormat(object):
    """Near Cache in memory format of the values."""

    BINARY = 0
    """
    As Hazelcast serialized bytearray data.
    """

    OBJECT = 1
    """
    As the actual object.
    """


class SSLProtocol(object):
    """SSL protocol options.

    TLSv1+ requires at least Python 2.7.9 or Python 3.4 build with OpenSSL 1.0.1+
    TLSv1_3 requires at least Python 2.7.15 or Python 3.7 build with OpenSSL 1.1.1+
    """

    SSLv2 = 0
    """
    SSL 2.0 Protocol. RFC 6176 prohibits SSL 2.0. Please use TLSv1+.
    """

    SSLv3 = 1
    """
    SSL 3.0 Protocol. RFC 7568 prohibits SSL 3.0. Please use TLSv1+.
    """

    TLSv1 = 2
    """
    TLS 1.0 Protocol described in RFC 2246.
    """

    TLSv1_1 = 3
    """
    TLS 1.1 Protocol described in RFC 4346.
    """

    TLSv1_2 = 4
    """
    TLS 1.2 Protocol described in RFC 5246.
    """

    TLSv1_3 = 5
    """
    TLS 1.3 Protocol described in RFC 8446.
    """


class QueryConstants(object):
    """Contains constants for Query."""

    KEY_ATTRIBUTE_NAME = "__key"
    """
    Attribute name of the key.
    """

    THIS_ATTRIBUTE_NAME = "this"
    """
    Attribute name of the value.
    """


class UniqueKeyTransformation(object):
    """Defines an assortment of transformations which can be applied to unique key values."""

    OBJECT = 0
    """
    Extracted unique key value is interpreted as an object value. 
    Non-negative unique ID is assigned to every distinct object value.
    """

    LONG = 1
    """
    Extracted unique key value is interpreted as a whole integer value of byte, short, int or long type. 
    The extracted value is up casted to long (if necessary) and unique non-negative ID is assigned 
    to every distinct value.
    """

    RAW = 2
    """
    Extracted unique key value is interpreted as a whole integer value of byte, short, int or long type. 
    The extracted value is up casted to long (if necessary) and the resulting value is used directly as an ID.
    """


class IndexType(object):
    """Type of the index."""

    SORTED = 0
    """
    Sorted index. Can be used with equality and range predicates.
    """

    HASH = 1
    """
    Hash index. Can be used with equality predicates.
    """

    BITMAP = 2
    """
    Bitmap index. Can be used with equality predicates.
    """


class ReconnectMode(object):
    """Reconnect options."""

    OFF = 0
    """
    Prevent reconnect to cluster after a disconnect.
    """

    ON = 1
    """
    Reconnect to cluster by blocking invocations.
    """

    ASYNC = 2
    """
    Reconnect to cluster without blocking invocations. Invocations will receive ClientOfflineError
    """


class TopicOverloadPolicy(object):
    """A policy to deal with an overloaded topic; a topic where there is no
    place to store new messages.

    The reliable topic uses a :class:`hazelcast.proxy.ringbuffer.Ringbuffer` to
    store the messages. A ringbuffer doesn't track where readers are, so
    it has no concept of a slow consumers. This provides many advantages like
    high performance reads, but it also gives the ability to the reader to
    re-read the same message multiple times in case of an error.

    A ringbuffer has a limited, fixed capacity. A fast producer may overwrite
    old messages that are still being read by a slow consumer. To prevent
    this, we may configure a time-to-live on the ringbuffer.

    Once the time-to-live is configured, the :class:`TopicOverloadPolicy`
    controls how the publisher is going to deal with the situation that a
    ringbuffer is full and the oldest item in the ringbuffer is not old
    enough to get overwritten.

    Keep in mind that this retention period (time-to-live) can keep messages
    from being overwritten, even though all readers might have already completed
    reading.
    """

    DISCARD_OLDEST = 0
    """Using this policy, a message that has not expired can be overwritten.
    
    No matter the retention period set, the overwrite will just overwrite
    the item.
    
    This can be a problem for slow consumers because they were promised a
    certain time window to process messages. But it will benefit producers
    and fast consumers since they are able to continue. This policy sacrifices
    the slow producer in favor of fast producers/consumers.
    """

    DISCARD_NEWEST = 1
    """The message that was to be published is discarded."""

    BLOCK = 2
    """The caller will wait till there space in the ringbuffer."""

    ERROR = 3
    """The publish call immediately fails."""


class BitmapIndexOptions(object):
    __slots__ = ("_unique_key", "_unique_key_transformation")

    def __init__(self, unique_key=None, unique_key_transformation=None):
        self._unique_key = QueryConstants.KEY_ATTRIBUTE_NAME
        if unique_key is not None:
            self.unique_key = unique_key

        self._unique_key_transformation = UniqueKeyTransformation.OBJECT
        if unique_key_transformation is not None:
            self.unique_key_transformation = unique_key_transformation

    @property
    def unique_key(self):
        return self._unique_key

    @unique_key.setter
    def unique_key(self, value):
        self._unique_key = try_to_get_enum_value(value, QueryConstants)

    @property
    def unique_key_transformation(self):
        return self._unique_key_transformation

    @unique_key_transformation.setter
    def unique_key_transformation(self, value):
        self._unique_key_transformation = try_to_get_enum_value(value, UniqueKeyTransformation)

    @classmethod
    def from_dict(cls, d):
        options = cls()
        for k, v in six.iteritems(d):
            try:
                options.__setattr__(k, v)
            except AttributeError:
                raise InvalidConfigurationError(
                    "Unrecognized config option for the bitmap index options: %s" % k
                )
        return options

    def __repr__(self):
        return "BitmapIndexOptions(unique_key=%s, unique_key_transformation=%s)" % (
            self.unique_key,
            self.unique_key_transformation,
        )


class IndexConfig(object):
    __slots__ = ("_name", "_type", "_attributes", "_bitmap_index_options")

    def __init__(self, name=None, type=None, attributes=None, bitmap_index_options=None):
        self._name = name
        if name is not None:
            self.name = name

        self._type = IndexType.SORTED
        if type is not None:
            self.type = type

        self._attributes = []
        if attributes is not None:
            self.attributes = attributes

        self._bitmap_index_options = BitmapIndexOptions()
        if bitmap_index_options is not None:
            self.bitmap_index_options = bitmap_index_options

    def add_attribute(self, attribute):
        IndexUtil.validate_attribute(attribute)
        self.attributes.append(attribute)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if isinstance(value, (six.string_types, none_type)):
            self._name = value
        else:
            raise TypeError("name must be a string or None")

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        self._type = try_to_get_enum_value(value, IndexType)

    @property
    def attributes(self):
        return self._attributes

    @attributes.setter
    def attributes(self, value):
        if isinstance(value, list):
            for attribute in value:
                IndexUtil.validate_attribute(attribute)
            self._attributes = value
        else:
            raise TypeError("attributes must be a list")

    @property
    def bitmap_index_options(self):
        return self._bitmap_index_options

    @bitmap_index_options.setter
    def bitmap_index_options(self, value):
        if isinstance(value, dict):
            self._bitmap_index_options = BitmapIndexOptions.from_dict(value)
        elif isinstance(value, BitmapIndexOptions):
            # This branch should only be taken by the client protocol
            self._bitmap_index_options = value
        else:
            raise TypeError("bitmap_index_options must be a dict")

    @classmethod
    def from_dict(cls, d):
        config = cls()
        for k, v in six.iteritems(d):
            if v is not None:
                try:
                    config.__setattr__(k, v)
                except AttributeError:
                    raise InvalidConfigurationError(
                        "Unrecognized config option for the index config: %s" % k
                    )
        return config

    def __repr__(self):
        return "IndexConfig(name=%s, type=%s, attributes=%s, bitmap_index_options=%s)" % (
            self.name,
            self.type,
            self.attributes,
            self.bitmap_index_options,
        )


class IndexUtil(object):
    _MAX_ATTRIBUTES = 255
    """Maximum number of attributes allowed in the index."""

    _THIS_PATTERN = re.compile(r"^this\.")
    """Pattern to stripe away "this." prefix."""

    @staticmethod
    def validate_attribute(attribute):
        check_not_none(attribute, "Attribute name cannot be None")

        stripped_attribute = attribute.strip()
        if not stripped_attribute:
            raise ValueError("Attribute name cannot be empty")

        if stripped_attribute.endswith("."):
            raise ValueError("Attribute name cannot end with dot: %s" % attribute)

    @staticmethod
    def validate_and_normalize(map_name, index_config):
        original_attributes = index_config.attributes
        if not original_attributes:
            raise ValueError("Index must have at least one attribute: %s" % index_config)

        if len(original_attributes) > IndexUtil._MAX_ATTRIBUTES:
            raise ValueError(
                "Index cannot have more than %s attributes %s"
                % (IndexUtil._MAX_ATTRIBUTES, index_config)
            )

        if index_config.type == IndexType.BITMAP and len(original_attributes) > 1:
            raise ValueError("Composite bitmap indexes are not supported: %s" % index_config)

        normalized_attributes = []
        for original_attribute in original_attributes:
            IndexUtil.validate_attribute(original_attribute)

            original_attribute = original_attribute.strip()
            normalized_attribute = IndexUtil.canonicalize_attribute(original_attribute)

            try:
                idx = normalized_attributes.index(normalized_attribute)
            except ValueError:
                pass
            else:
                duplicate_original_attribute = original_attributes[idx]
                if duplicate_original_attribute == original_attribute:
                    raise ValueError(
                        "Duplicate attribute name [attribute_name=%s, index_config=%s]"
                        % (original_attribute, index_config)
                    )
                else:
                    raise ValueError(
                        "Duplicate attribute names [attribute_name1=%s, attribute_name2=%s, "
                        "index_config=%s]"
                        % (duplicate_original_attribute, original_attribute, index_config)
                    )

            normalized_attributes.append(normalized_attribute)

        name = index_config.name
        if name and not name.strip():
            name = None

        normalized_config = IndexUtil.build_normalized_config(
            map_name, index_config.type, name, normalized_attributes
        )
        if index_config.type == IndexType.BITMAP:
            unique_key = index_config.bitmap_index_options.unique_key
            unique_key_transformation = index_config.bitmap_index_options.unique_key_transformation
            IndexUtil.validate_attribute(unique_key)
            unique_key = IndexUtil.canonicalize_attribute(unique_key)
            normalized_config.bitmap_index_options.unique_key = unique_key
            normalized_config.bitmap_index_options.unique_key_transformation = (
                unique_key_transformation
            )

        return normalized_config

    @staticmethod
    def canonicalize_attribute(attribute):
        return re.sub(IndexUtil._THIS_PATTERN, "", attribute)

    @staticmethod
    def build_normalized_config(map_name, index_type, index_name, normalized_attributes):
        new_config = IndexConfig()
        new_config.type = index_type

        name = (
            map_name + "_" + IndexUtil._index_type_to_name(index_type)
            if index_name is None
            else None
        )
        for normalized_attribute in normalized_attributes:
            new_config.add_attribute(normalized_attribute)
            if name:
                name += "_" + normalized_attribute

        if name:
            index_name = name

        new_config.name = index_name
        return new_config

    @staticmethod
    def _index_type_to_name(index_type):
        if index_type == IndexType.SORTED:
            return "sorted"
        elif index_type == IndexType.HASH:
            return "hash"
        elif index_type == IndexType.BITMAP:
            return "bitmap"
        else:
            raise ValueError("Unsupported index type %s" % index_type)


_DEFAULT_CLUSTER_NAME = "dev"
_DEFAULT_CONNECTION_TIMEOUT = 5.0
_DEFAULT_RETRY_INITIAL_BACKOFF = 1.0
_DEFAULT_RETRY_MAX_BACKOFF = 30.0
_DEFAULT_RETRY_JITTER = 0.0
_DEFAULT_RETRY_MULTIPLIER = 1.05
_DEFAULT_CLUSTER_CONNECT_TIMEOUT = -1
_DEFAULT_PORTABLE_VERSION = 0
_DEFAULT_HEARTBEAT_INTERVAL = 5.0
_DEFAULT_HEARTBEAT_TIMEOUT = 60.0
_DEFAULT_INVOCATION_TIMEOUT = 120.0
_DEFAULT_INVOCATION_RETRY_PAUSE = 1.0
_DEFAULT_STATISTICS_PERIOD = 3.0
_DEFAULT_OPERATION_BACKUP_TIMEOUT = 5.0


class _Config(object):
    __slots__ = (
        "_cluster_members",
        "_cluster_name",
        "_client_name",
        "_connection_timeout",
        "_socket_options",
        "_redo_operation",
        "_smart_routing",
        "_ssl_enabled",
        "_ssl_cafile",
        "_ssl_certfile",
        "_ssl_keyfile",
        "_ssl_password",
        "_ssl_protocol",
        "_ssl_ciphers",
        "_cloud_discovery_token",
        "_async_start",
        "_reconnect_mode",
        "_retry_initial_backoff",
        "_retry_max_backoff",
        "_retry_jitter",
        "_retry_multiplier",
        "_cluster_connect_timeout",
        "_portable_version",
        "_data_serializable_factories",
        "_portable_factories",
        "_class_definitions",
        "_check_class_definition_errors",
        "_is_big_endian",
        "_default_int_type",
        "_global_serializer",
        "_custom_serializers",
        "_near_caches",
        "_load_balancer",
        "_membership_listeners",
        "_lifecycle_listeners",
        "_flake_id_generators",
        "_reliable_topics",
        "_labels",
        "_heartbeat_interval",
        "_heartbeat_timeout",
        "_invocation_timeout",
        "_invocation_retry_pause",
        "_statistics_enabled",
        "_statistics_period",
        "_shuffle_member_list",
        "_backup_ack_to_client_enabled",
        "_operation_backup_timeout",
        "_fail_on_indeterminate_operation_state",
        "_creds_username",
        "_creds_password",
        "_token_provider",
    )

    def __init__(self):
        self._cluster_members = []
        self._cluster_name = _DEFAULT_CLUSTER_NAME
        self._client_name = None
        self._connection_timeout = _DEFAULT_CONNECTION_TIMEOUT
        self._socket_options = []
        self._redo_operation = False
        self._smart_routing = True
        self._ssl_enabled = False
        self._ssl_cafile = None
        self._ssl_certfile = None
        self._ssl_keyfile = None
        self._ssl_password = None
        self._ssl_protocol = SSLProtocol.TLSv1_2
        self._ssl_ciphers = None
        self._cloud_discovery_token = None
        self._async_start = False
        self._reconnect_mode = ReconnectMode.ON
        self._retry_initial_backoff = _DEFAULT_RETRY_INITIAL_BACKOFF
        self._retry_max_backoff = _DEFAULT_RETRY_MAX_BACKOFF
        self._retry_jitter = _DEFAULT_RETRY_JITTER
        self._retry_multiplier = _DEFAULT_RETRY_MULTIPLIER
        self._cluster_connect_timeout = _DEFAULT_CLUSTER_CONNECT_TIMEOUT
        self._portable_version = _DEFAULT_PORTABLE_VERSION
        self._data_serializable_factories = {}
        self._portable_factories = {}
        self._class_definitions = []
        self._check_class_definition_errors = True
        self._is_big_endian = True
        self._default_int_type = IntType.INT
        self._global_serializer = None
        self._custom_serializers = {}
        self._near_caches = {}
        self._load_balancer = None
        self._membership_listeners = []
        self._lifecycle_listeners = []
        self._flake_id_generators = {}
        self._reliable_topics = {}
        self._labels = []
        self._heartbeat_interval = _DEFAULT_HEARTBEAT_INTERVAL
        self._heartbeat_timeout = _DEFAULT_HEARTBEAT_TIMEOUT
        self._invocation_timeout = _DEFAULT_INVOCATION_TIMEOUT
        self._invocation_retry_pause = _DEFAULT_INVOCATION_RETRY_PAUSE
        self._statistics_enabled = False
        self._statistics_period = _DEFAULT_STATISTICS_PERIOD
        self._shuffle_member_list = True
        self._backup_ack_to_client_enabled = True
        self._operation_backup_timeout = _DEFAULT_OPERATION_BACKUP_TIMEOUT
        self._fail_on_indeterminate_operation_state = False
        self._creds_username = None
        self._creds_password = None
        self._token_provider = None

    @property
    def cluster_members(self):
        return self._cluster_members

    @cluster_members.setter
    def cluster_members(self, value):
        if isinstance(value, list):
            for address in value:
                if not isinstance(address, six.string_types):
                    raise TypeError("cluster_members must be list of strings")

            self._cluster_members = value
        else:
            raise TypeError("cluster_members must be a list")

    @property
    def cluster_name(self):
        return self._cluster_name

    @cluster_name.setter
    def cluster_name(self, value):
        if isinstance(value, six.string_types):
            self._cluster_name = value
        else:
            raise TypeError("cluster_name must be a string")

    @property
    def client_name(self):
        return self._client_name

    @client_name.setter
    def client_name(self, value):
        if isinstance(value, six.string_types):
            self._client_name = value
        else:
            raise TypeError("client_name must be a string")

    @property
    def connection_timeout(self):
        return self._connection_timeout

    @connection_timeout.setter
    def connection_timeout(self, value):
        if isinstance(value, number_types):
            if value < 0:
                raise ValueError("connection_timeout must be non-negative")
            self._connection_timeout = value
        else:
            raise TypeError("connection_timeout must be a number")

    @property
    def socket_options(self):
        return self._socket_options

    @socket_options.setter
    def socket_options(self, value):
        if isinstance(value, list):
            try:
                for _, _, _ in value:
                    # Must be a tuple of length 3
                    pass

                self._socket_options = value
            except ValueError:
                raise TypeError("socket_options must contain tuples of length 3 as items")
        else:
            raise TypeError("socket_options must be a list")

    @property
    def redo_operation(self):
        return self._redo_operation

    @redo_operation.setter
    def redo_operation(self, value):
        if isinstance(value, bool):
            self._redo_operation = value
        else:
            raise TypeError("redo_operation must be a boolean")

    @property
    def smart_routing(self):
        return self._smart_routing

    @smart_routing.setter
    def smart_routing(self, value):
        if isinstance(value, bool):
            self._smart_routing = value
        else:
            raise TypeError("smart_routing must be a boolean")

    @property
    def ssl_enabled(self):
        return self._ssl_enabled

    @ssl_enabled.setter
    def ssl_enabled(self, value):
        if isinstance(value, bool):
            self._ssl_enabled = value
        else:
            raise TypeError("ssl_enabled must be a boolean")

    @property
    def ssl_cafile(self):
        return self._ssl_cafile

    @ssl_cafile.setter
    def ssl_cafile(self, value):
        if isinstance(value, six.string_types):
            self._ssl_cafile = value
        else:
            raise TypeError("ssl_cafile must be a string")

    @property
    def ssl_certfile(self):
        return self._ssl_certfile

    @ssl_certfile.setter
    def ssl_certfile(self, value):
        if isinstance(value, six.string_types):
            self._ssl_certfile = value
        else:
            raise TypeError("ssl_certfile must be a string")

    @property
    def ssl_keyfile(self):
        return self._ssl_keyfile

    @ssl_keyfile.setter
    def ssl_keyfile(self, value):
        if isinstance(value, six.string_types):
            self._ssl_keyfile = value
        else:
            raise TypeError("ssl_keyfile must be a string")

    @property
    def ssl_password(self):
        return self._ssl_password

    @ssl_password.setter
    def ssl_password(self, value):
        if isinstance(value, (six.string_types, six.binary_type, bytearray)) or callable(value):
            self._ssl_password = value
        else:
            raise TypeError("ssl_password must be string, bytes, bytearray or callable")

    @property
    def ssl_protocol(self):
        return self._ssl_protocol

    @ssl_protocol.setter
    def ssl_protocol(self, value):
        self._ssl_protocol = try_to_get_enum_value(value, SSLProtocol)

    @property
    def ssl_ciphers(self):
        return self._ssl_ciphers

    @ssl_ciphers.setter
    def ssl_ciphers(self, value):
        if isinstance(value, six.string_types):
            self._ssl_ciphers = value
        else:
            raise TypeError("ssl_ciphers must be a string")

    @property
    def cloud_discovery_token(self):
        return self._cloud_discovery_token

    @cloud_discovery_token.setter
    def cloud_discovery_token(self, value):
        if isinstance(value, six.string_types):
            self._cloud_discovery_token = value
        else:
            raise TypeError("cloud_discovery_token must be a string")

    @property
    def async_start(self):
        return self._async_start

    @async_start.setter
    def async_start(self, value):
        if isinstance(value, bool):
            self._async_start = value
        else:
            raise TypeError("async_start must be a boolean")

    @property
    def reconnect_mode(self):
        return self._reconnect_mode

    @reconnect_mode.setter
    def reconnect_mode(self, value):
        self._reconnect_mode = try_to_get_enum_value(value, ReconnectMode)

    @property
    def retry_initial_backoff(self):
        return self._retry_initial_backoff

    @retry_initial_backoff.setter
    def retry_initial_backoff(self, value):
        if isinstance(value, number_types):
            if value < 0:
                raise ValueError("retry_initial_backoff must be non-negative")
            self._retry_initial_backoff = value
        else:
            raise TypeError("retry_initial_backoff must be a number")

    @property
    def retry_max_backoff(self):
        return self._retry_max_backoff

    @retry_max_backoff.setter
    def retry_max_backoff(self, value):
        if isinstance(value, number_types):
            if value < 0:
                raise ValueError("retry_max_backoff must be non-negative")
            self._retry_max_backoff = value
        else:
            raise TypeError("retry_max_backoff must be a number")

    @property
    def retry_jitter(self):
        return self._retry_jitter

    @retry_jitter.setter
    def retry_jitter(self, value):
        if isinstance(value, number_types):
            if value < 0 or value > 1:
                raise ValueError("retry_jitter must be in range [0.0, 1.0]")
            self._retry_jitter = value
        else:
            raise TypeError("retry_jitter must be a number")

    @property
    def retry_multiplier(self):
        return self._retry_multiplier

    @retry_multiplier.setter
    def retry_multiplier(self, value):
        if isinstance(value, number_types):
            if value < 1:
                raise ValueError("retry_multiplier must be greater than or equal to 1.0")
            self._retry_multiplier = value
        else:
            raise TypeError("retry_multiplier must be a number")

    @property
    def cluster_connect_timeout(self):
        return self._cluster_connect_timeout

    @cluster_connect_timeout.setter
    def cluster_connect_timeout(self, value):
        if isinstance(value, number_types):
            if value < 0 and value != _DEFAULT_CLUSTER_CONNECT_TIMEOUT:
                raise ValueError(
                    "cluster_connect_timeout must be non-negative or equal to %s"
                    % _DEFAULT_CLUSTER_CONNECT_TIMEOUT
                )
            self._cluster_connect_timeout = value
        else:
            raise TypeError("cluster_connect_timeout must be a number")

    @property
    def portable_version(self):
        return self._portable_version

    @portable_version.setter
    def portable_version(self, value):
        if isinstance(value, number_types):
            if value < 0:
                raise ValueError("portable_version must be non-negative")
            self._portable_version = value
        else:
            raise TypeError("portable_version must be a number")

    @property
    def data_serializable_factories(self):
        return self._data_serializable_factories

    @data_serializable_factories.setter
    def data_serializable_factories(self, value):
        if isinstance(value, dict):
            for factory_id, factory in six.iteritems(value):
                if not isinstance(factory_id, six.integer_types):
                    raise TypeError("Keys of data_serializable_factories must be integers")

                if not isinstance(factory, dict):
                    raise TypeError("Values of data_serializable_factories must be dict")

                for class_id, clazz in six.iteritems(factory):
                    if not isinstance(class_id, six.integer_types):
                        raise TypeError(
                            "Keys of factories of data_serializable_factories must be integers"
                        )

                    if not (
                        isinstance(clazz, type) and issubclass(clazz, IdentifiedDataSerializable)
                    ):
                        raise TypeError(
                            "Values of factories of data_serializable_factories must be "
                            "subclasses of IdentifiedDataSerializable"
                        )

            self._data_serializable_factories = value
        else:
            raise TypeError("data_serializable_factories must be a dict")

    @property
    def portable_factories(self):
        return self._portable_factories

    @portable_factories.setter
    def portable_factories(self, value):
        if isinstance(value, dict):
            for factory_id, factory in six.iteritems(value):
                if not isinstance(factory_id, six.integer_types):
                    raise TypeError("Keys of portable_factories must be integers")

                if not isinstance(factory, dict):
                    raise TypeError("Values of portable_factories must be dict")

                for class_id, clazz in six.iteritems(factory):
                    if not isinstance(class_id, six.integer_types):
                        raise TypeError("Keys of factories of portable_factories must be integers")

                    if not (isinstance(clazz, type) and issubclass(clazz, Portable)):
                        raise TypeError(
                            "Values of factories of portable_factories must be "
                            "subclasses of Portable"
                        )

            self._portable_factories = value
        else:
            raise TypeError("portable_factories must be a dict")

    @property
    def class_definitions(self):
        return self._class_definitions

    @class_definitions.setter
    def class_definitions(self, value):
        if isinstance(value, list):
            for cd in value:
                if not isinstance(cd, ClassDefinition):
                    raise TypeError(
                        "class_definitions must contain objects of type ClassDefinition"
                    )

            self._class_definitions = value
        else:
            raise TypeError("class_definitions must be a list")

    @property
    def check_class_definition_errors(self):
        return self._check_class_definition_errors

    @check_class_definition_errors.setter
    def check_class_definition_errors(self, value):
        if isinstance(value, bool):
            self._check_class_definition_errors = value
        else:
            raise TypeError("check_class_definition_errors must be a boolean")

    @property
    def is_big_endian(self):
        return self._is_big_endian

    @is_big_endian.setter
    def is_big_endian(self, value):
        if isinstance(value, bool):
            self._is_big_endian = value
        else:
            raise TypeError("is_big_endian must be a boolean")

    @property
    def default_int_type(self):
        return self._default_int_type

    @default_int_type.setter
    def default_int_type(self, value):
        self._default_int_type = try_to_get_enum_value(value, IntType)

    @property
    def global_serializer(self):
        return self._global_serializer

    @global_serializer.setter
    def global_serializer(self, value):
        if isinstance(value, type) and issubclass(value, StreamSerializer):
            self._global_serializer = value
        else:
            raise TypeError("global_serializer must be a StreamSerializer")

    @property
    def custom_serializers(self):
        return self._custom_serializers

    @custom_serializers.setter
    def custom_serializers(self, value):
        if isinstance(value, dict):
            for _type, serializer in six.iteritems(value):
                if not isinstance(_type, type):
                    raise TypeError("Keys of custom_serializers must be types")

                if not (isinstance(serializer, type) and issubclass(serializer, StreamSerializer)):
                    raise TypeError(
                        "Values of custom_serializers must be subclasses of StreamSerializer"
                    )

            self._custom_serializers = value
        else:
            raise TypeError("custom_serializers must be a dict")

    @property
    def near_caches(self):
        return self._near_caches

    @near_caches.setter
    def near_caches(self, value):
        if isinstance(value, dict):
            configs = dict()
            for name, config in six.iteritems(value):
                if not isinstance(name, six.string_types):
                    raise TypeError("Keys of near_caches must be strings")

                if not isinstance(config, dict):
                    raise TypeError("Values of near_caches must be dict")

                configs[name] = _NearCacheConfig.from_dict(config)

            self._near_caches = configs
        else:
            raise TypeError("near_caches must be a dict")

    @property
    def load_balancer(self):
        return self._load_balancer

    @load_balancer.setter
    def load_balancer(self, value):
        if isinstance(value, LoadBalancer):
            self._load_balancer = value
        else:
            raise TypeError("load_balancer must be a LoadBalancer")

    @property
    def membership_listeners(self):
        return self._membership_listeners

    @membership_listeners.setter
    def membership_listeners(self, value):
        if isinstance(value, list):
            try:
                for item in value:
                    try:
                        added, removed = item
                    except TypeError:
                        raise TypeError(
                            "membership_listeners must contain tuples of length 2 as items"
                        )

                    if not (callable(added) or callable(removed)):
                        raise TypeError(
                            "At least one of the listeners in the tuple most be callable"
                        )

                self._membership_listeners = value
            except ValueError:
                raise TypeError("membership_listeners must contain tuples of length 2 as items")
        else:
            raise TypeError("membership_listeners must be a list")

    @property
    def lifecycle_listeners(self):
        return self._lifecycle_listeners

    @lifecycle_listeners.setter
    def lifecycle_listeners(self, value):
        if isinstance(value, list):
            for listener in value:
                if not callable(listener):
                    raise TypeError("lifecycle_listeners must contain callable items")

            self._lifecycle_listeners = value
        else:
            raise TypeError("lifecycle_listeners must be a list")

    @property
    def flake_id_generators(self):
        return self._flake_id_generators

    @flake_id_generators.setter
    def flake_id_generators(self, value):
        if isinstance(value, dict):
            configs = dict()
            for name, config in six.iteritems(value):
                if not isinstance(name, six.string_types):
                    raise TypeError("Keys of flake_id_generators must be strings")

                if not isinstance(config, dict):
                    raise TypeError("Values of flake_id_generators must be dict")

                configs[name] = _FlakeIdGeneratorConfig.from_dict(config)

            self._flake_id_generators = configs
        else:
            raise TypeError("flake_id_generators must be a dict")

    @property
    def reliable_topics(self):
        return self._reliable_topics

    @reliable_topics.setter
    def reliable_topics(self, value):
        if isinstance(value, dict):
            configs = {}
            for name, config in six.iteritems(value):
                if not isinstance(name, six.string_types):
                    raise TypeError("Keys of reliable_topics must be strings")

                if not isinstance(config, dict):
                    raise TypeError("Values of reliable_topics must be dict")

                configs[name] = _ReliableTopicConfig.from_dict(config)

            self._reliable_topics = configs
        else:
            raise TypeError("reliable_topics must be a dict")

    @property
    def labels(self):
        return self._labels

    @labels.setter
    def labels(self, value):
        if isinstance(value, list):
            for label in value:
                if not isinstance(label, six.string_types):
                    raise TypeError("labels must be list of strings")

            self._labels = value
        else:
            raise TypeError("labels must be a list")

    @property
    def heartbeat_interval(self):
        return self._heartbeat_interval

    @heartbeat_interval.setter
    def heartbeat_interval(self, value):
        if isinstance(value, number_types):
            if value <= 0:
                raise ValueError("heartbeat_interval must be positive")
            self._heartbeat_interval = value
        else:
            raise TypeError("heartbeat_interval must be a number")

    @property
    def heartbeat_timeout(self):
        return self._heartbeat_timeout

    @heartbeat_timeout.setter
    def heartbeat_timeout(self, value):
        if isinstance(value, number_types):
            if value <= 0:
                raise ValueError("heartbeat_timeout must be positive")
            self._heartbeat_timeout = value
        else:
            raise TypeError("heartbeat_timeout must be a number")

    @property
    def invocation_timeout(self):
        return self._invocation_timeout

    @invocation_timeout.setter
    def invocation_timeout(self, value):
        if isinstance(value, number_types):
            if value <= 0:
                raise ValueError("invocation_timeout must be positive")
            self._invocation_timeout = value
        else:
            raise TypeError("invocation_timeout must be a number")

    @property
    def invocation_retry_pause(self):
        return self._invocation_retry_pause

    @invocation_retry_pause.setter
    def invocation_retry_pause(self, value):
        if isinstance(value, number_types):
            if value <= 0:
                raise ValueError("invocation_retry_pause must be positive")
            self._invocation_retry_pause = value
        else:
            raise TypeError("invocation_retry_pause must be a number")

    @property
    def statistics_enabled(self):
        return self._statistics_enabled

    @statistics_enabled.setter
    def statistics_enabled(self, value):
        if isinstance(value, bool):
            self._statistics_enabled = value
        else:
            raise TypeError("statistics_enabled must be a boolean")

    @property
    def statistics_period(self):
        return self._statistics_period

    @statistics_period.setter
    def statistics_period(self, value):
        if isinstance(value, number_types):
            if value <= 0:
                raise ValueError("statistics_period must be positive")
            self._statistics_period = value
        else:
            raise TypeError("statistics_period must be a number")

    @property
    def shuffle_member_list(self):
        return self._shuffle_member_list

    @shuffle_member_list.setter
    def shuffle_member_list(self, value):
        if isinstance(value, bool):
            self._shuffle_member_list = value
        else:
            raise TypeError("shuffle_member_list must be a boolean")

    @property
    def backup_ack_to_client_enabled(self):
        return self._backup_ack_to_client_enabled

    @backup_ack_to_client_enabled.setter
    def backup_ack_to_client_enabled(self, value):
        if isinstance(value, bool):
            self._backup_ack_to_client_enabled = value
        else:
            raise TypeError("backup_ack_to_client_enabled must be a boolean")

    @property
    def operation_backup_timeout(self):
        return self._operation_backup_timeout

    @operation_backup_timeout.setter
    def operation_backup_timeout(self, value):
        if isinstance(value, number_types):
            if value > 0:
                self._operation_backup_timeout = value
            else:
                raise ValueError("operation_backup_timeout must be positive")
        else:
            raise TypeError("operation_backup_timeout must be a number")

    @property
    def fail_on_indeterminate_operation_state(self):
        return self._fail_on_indeterminate_operation_state

    @fail_on_indeterminate_operation_state.setter
    def fail_on_indeterminate_operation_state(self, value):
        if isinstance(value, bool):
            self._fail_on_indeterminate_operation_state = value
        else:
            raise TypeError("fail_on_indeterminate_operation_state must be a boolean")

    @property
    def creds_username(self):
        # type: (_Config) -> str
        return self._creds_username

    @creds_username.setter
    def creds_username(self, username):
        # type: (_Config, str) -> None
        if not isinstance(username, six.string_types):
            raise TypeError("creds_password must be a string")
        self._creds_username = username

    @property
    def creds_password(self):
        # type: (_Config) -> str
        return self._creds_password

    @creds_password.setter
    def creds_password(self, password):
        # type: (_Config, str) -> None
        if not isinstance(password, six.string_types):
            raise TypeError("creds_password must be a string")
        self._creds_password = password

    @property
    def token_provider(self):
        # type: (_Config) -> TokenProvider
        return self._token_provider

    @token_provider.setter
    def token_provider(self, token_provider):
        # type: (_Config, TokenProvider) -> None
        token_fun = getattr(token_provider, "token", None)
        if token_fun is None or not isinstance(token_fun, types.MethodType):
            raise TypeError("token_provider must be an object with a token method")
        self._token_provider = token_provider

    @classmethod
    def from_dict(cls, d):
        config = cls()
        for k, v in six.iteritems(d):
            if v is not None:
                try:
                    config.__setattr__(k, v)
                except AttributeError:
                    raise InvalidConfigurationError("Unrecognized config option: %s" % k)
        return config


class _NearCacheConfig(object):
    __slots__ = (
        "_invalidate_on_change",
        "_in_memory_format",
        "_time_to_live",
        "_max_idle",
        "_eviction_policy",
        "_eviction_max_size",
        "_eviction_sampling_count",
        "_eviction_sampling_pool_size",
    )

    def __init__(self):
        self._invalidate_on_change = True
        self._in_memory_format = InMemoryFormat.BINARY
        self._time_to_live = None
        self._max_idle = None
        self._eviction_policy = EvictionPolicy.LRU
        self._eviction_max_size = 10000
        self._eviction_sampling_count = 8
        self._eviction_sampling_pool_size = 16

    @property
    def invalidate_on_change(self):
        return self._invalidate_on_change

    @invalidate_on_change.setter
    def invalidate_on_change(self, value):
        if isinstance(value, bool):
            self._invalidate_on_change = value
        else:
            raise TypeError("invalidate_on_change must be a boolean")

    @property
    def in_memory_format(self):
        return self._in_memory_format

    @in_memory_format.setter
    def in_memory_format(self, value):
        self._in_memory_format = try_to_get_enum_value(value, InMemoryFormat)

    @property
    def time_to_live(self):
        return self._time_to_live

    @time_to_live.setter
    def time_to_live(self, value):
        if isinstance(value, number_types):
            if value < 0:
                raise ValueError("time_to_live must be non-negative")
            self._time_to_live = value
        else:
            raise TypeError("time_to_live must be a number")

    @property
    def max_idle(self):
        return self._max_idle

    @max_idle.setter
    def max_idle(self, value):
        if isinstance(value, number_types):
            if value < 0:
                raise ValueError("max_idle must be non-negative")
            self._max_idle = value
        else:
            raise TypeError("max_idle must be a number")

    @property
    def eviction_policy(self):
        return self._eviction_policy

    @eviction_policy.setter
    def eviction_policy(self, value):
        self._eviction_policy = try_to_get_enum_value(value, EvictionPolicy)

    @property
    def eviction_max_size(self):
        return self._eviction_max_size

    @eviction_max_size.setter
    def eviction_max_size(self, value):
        if isinstance(value, number_types):
            if value < 1:
                raise ValueError("eviction_max_size must be greater than 1")
            self._eviction_max_size = value
        else:
            raise TypeError("eviction_max_size must be a number")

    @property
    def eviction_sampling_count(self):
        return self._eviction_sampling_count

    @eviction_sampling_count.setter
    def eviction_sampling_count(self, value):
        if isinstance(value, number_types):
            if value < 1:
                raise ValueError("eviction_sampling_count must be greater than 1")
            self._eviction_sampling_count = value
        else:
            raise TypeError("eviction_sampling_count must be a number")

    @property
    def eviction_sampling_pool_size(self):
        return self._eviction_sampling_pool_size

    @eviction_sampling_pool_size.setter
    def eviction_sampling_pool_size(self, value):
        if isinstance(value, number_types):
            if value < 1:
                raise ValueError("eviction_sampling_pool_size must be greater than 1")
            self._eviction_sampling_pool_size = value
        else:
            raise TypeError("eviction_sampling_pool_size must be a number")

    @classmethod
    def from_dict(cls, d):
        config = cls()
        for k, v in six.iteritems(d):
            try:
                config.__setattr__(k, v)
            except AttributeError:
                raise InvalidConfigurationError(
                    "Unrecognized config option for the near cache: %s" % k
                )
        return config


class _FlakeIdGeneratorConfig(object):
    __slots__ = ("_prefetch_count", "_prefetch_validity")

    def __init__(self):
        self._prefetch_count = 100
        self._prefetch_validity = 600

    @property
    def prefetch_count(self):
        return self._prefetch_count

    @prefetch_count.setter
    def prefetch_count(self, value):
        if isinstance(value, number_types):
            if not (0 < value <= 100000):
                raise ValueError("prefetch_count must be in range 1 to 100000")
            self._prefetch_count = value
        else:
            raise TypeError("prefetch_count must be a number")

    @property
    def prefetch_validity(self):
        return self._prefetch_validity

    @prefetch_validity.setter
    def prefetch_validity(self, value):
        if isinstance(value, number_types):
            if value < 0:
                raise ValueError("prefetch_validity must be non-negative")
            self._prefetch_validity = value
        else:
            raise TypeError("prefetch_validity must be a number")

    @classmethod
    def from_dict(cls, d):
        config = cls()
        for k, v in six.iteritems(d):
            try:
                config.__setattr__(k, v)
            except AttributeError:
                raise InvalidConfigurationError(
                    "Unrecognized config option for the flake id generator: %s" % k
                )
        return config


class _ReliableTopicConfig(object):
    __slots__ = ("_read_batch_size", "_overload_policy")

    def __init__(self):
        self._read_batch_size = 10
        self._overload_policy = TopicOverloadPolicy.BLOCK

    @property
    def read_batch_size(self):
        return self._read_batch_size

    @read_batch_size.setter
    def read_batch_size(self, value):
        if isinstance(value, number_types):
            if value <= 0:
                raise ValueError("read_batch_size must be positive")
            self._read_batch_size = value
        else:
            raise TypeError("read_batch_size must be a number")

    @property
    def overload_policy(self):
        return self._overload_policy

    @overload_policy.setter
    def overload_policy(self, value):
        self._overload_policy = try_to_get_enum_value(value, TopicOverloadPolicy)

    @classmethod
    def from_dict(cls, d):
        config = cls()
        for k, v in six.iteritems(d):
            try:
                config.__setattr__(k, v)
            except AttributeError:
                raise InvalidConfigurationError(
                    "Unrecognized config option for the reliable topic: %s" % k
                )
        return config



from hazelcast.protocol.codec import (
    topic_add_message_listener_codec,
    topic_publish_codec,
    topic_remove_message_listener_codec,
)
from hazelcast.proxy.base import PartitionSpecificProxy, TopicMessage


class Topic(PartitionSpecificProxy):
    """Hazelcast provides distribution mechanism for publishing messages that are delivered to multiple subscribers,
    which is also known as a publish/subscribe (pub/sub) messaging model.

    Publish and subscriptions are cluster-wide. When a member subscribes for a topic,
    it is actually registering for messages published by any member in the cluster,
    including the new members joined after you added the listener.

    Messages are ordered, meaning that listeners(subscribers) will process the messages in the order they are actually
    published.
    """

    def add_listener(self, on_message=None):
        """Subscribes to this topic.

        When someone publishes a message on this topic, ``on_message`` function is called if provided.

        Args:
            on_message (function): Function to be called when a message is published.

        Returns:
            hazelcast.future.Future[str]: A registration id which is used as a key to remove the listener.
        """
        codec = topic_add_message_listener_codec
        request = codec.encode_request(self.name, self._is_smart)

        def handle(item, publish_time, uuid):
            member = self._context.cluster_service.get_member(uuid)
            item_event = TopicMessage(
                self.name, item, publish_time / 1000.0, member, self._to_object
            )
            on_message(item_event)

        return self._register_listener(
            request,
            lambda r: codec.decode_response(r),
            lambda reg_id: topic_remove_message_listener_codec.encode_request(self.name, reg_id),
            lambda m: codec.handle(m, handle),
        )

    def publish(self, message):
        """Publishes the message to all subscribers of this topic

        Args:
            message: The message to be published.

        Returns:
            hazelcast.future.Future[None]:
        """
        message_data = self._to_data(message)
        request = topic_publish_codec.encode_request(self.name, message_data)
        return self._invoke(request)

    def remove_listener(self, registration_id):
        """Stops receiving messages for the given message listener.

        If the given listener already removed, this method does nothing.

        Args:
            registration_id (str): Registration id of the listener to be removed.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the listener is removed, ``False`` otherwise.
        """
        return self._deregister_listener(registration_id)



import time
import uuid

from hazelcast.errors import (
    LockOwnershipLostError,
    LockAcquireLimitReachedError,
    SessionExpiredError,
    WaitKeyCancelledError,
    IllegalMonitorStateError,
)
from hazelcast.future import ImmediateExceptionFuture
from hazelcast.protocol.codec import (
    fenced_lock_lock_codec,
    fenced_lock_try_lock_codec,
    fenced_lock_unlock_codec,
    fenced_lock_get_lock_ownership_codec,
)
from hazelcast.proxy.cp import SessionAwareCPProxy
from hazelcast.util import thread_id, to_millis

_NO_SESSION_ID = -1


class FencedLock(SessionAwareCPProxy):
    """A linearizable, distributed lock.

    FencedLock is CP with respect to the CAP principle. It works on top
    of the Raft consensus algorithm. It offers linearizability during crash-stop
    failures and network partitions. If a network partition occurs, it remains
    available on at most one side of the partition.

    FencedLock works on top of CP sessions. Please refer to CP Session
    IMDG documentation section for more information.

    By default, FencedLock is reentrant. Once a caller acquires
    the lock, it can acquire the lock reentrantly as many times as it wants
    in a linearizable manner. You can configure the reentrancy behaviour
    on the member side. For instance, reentrancy can be disabled and
    FencedLock can work as a non-reentrant mutex. One can also set
    a custom reentrancy limit. When the reentrancy limit is reached,
    FencedLock does not block a lock call. Instead, it fails with
    ``LockAcquireLimitReachedError`` or a specified return value.
    Please check the locking methods to see details about the behaviour.

    It is advised to use this proxy in a blocking mode. Although it is
    possible, non-blocking usage requires an extra care. FencedLock
    uses the id of the thread that makes the request to distinguish lock
    owners. When used in a non-blocking mode, added callbacks or
    continuations are not generally executed in the thread that makes the
    request. That causes the code below to fail most of the time since the
    lock is acquired on the main thread but, unlock request is done in another
    thread. ::

        lock = client.cp_subsystem.get_lock("lock")

        def cb(_):
            lock.unlock()

        lock.lock().add_done_callback(cb)
    """

    INVALID_FENCE = 0

    def __init__(self, context, group_id, service_name, proxy_name, object_name):
        super(FencedLock, self).__init__(context, group_id, service_name, proxy_name, object_name)
        self._lock_session_ids = dict()  # thread-id to session id that has acquired the lock

    def lock(self):
        """Acquires the lock and returns the fencing token assigned to the current
        thread for this lock acquire.

        If the lock is acquired reentrantly, the same fencing token is returned,
        or the ``lock()`` call can fail with ``LockAcquireLimitReachedError``
        if the lock acquire limit is already reached.

        If the lock is not available then the current thread becomes disabled
        for thread scheduling purposes and lies dormant until the lock has been
        acquired.

        Fencing tokens are monotonic numbers that are incremented each time
        the lock switches from the free state to the acquired state. They are
        simply used for ordering lock holders. A lock holder can pass
        its fencing to the shared resource to fence off previous lock holders.
        When this resource receives an operation, it can validate the fencing
        token in the operation.

        Consider the following scenario where the lock is free initially ::

            lock = client.cp_subsystem.get_lock("lock").blocking()
            fence1 = lock.lock()  # (1)
            fence2 = lock.lock()  # (2)
            assert fence1 == fence2
            lock.unlock()
            lock.unlock()
            fence3 = lock.lock()  # (3)
            assert fence3 > fence1

        In this scenario, the lock is acquired by a thread in the cluster. Then,
        the same thread reentrantly acquires the lock again. The fencing token
        returned from the second acquire is equal to the one returned from the
        first acquire, because of reentrancy. After the second acquire, the lock
        is released 2 times, hence becomes free. There is a third lock acquire
        here, which returns a new fencing token. Because this last lock acquire
        is not reentrant, its fencing token is guaranteed to be larger than the
        previous tokens, independent of the thread that has acquired the lock.

        Returns:
            hazelcast.future.Future[int]: The fencing token.

        Raises:
            LockOwnershipLostError: If the underlying CP session was
                closed before the client releases the lock
            LockAcquireLimitReachedError: If the lock call is reentrant
                and the configured lock acquire limit is already reached.
        """
        current_thread_id = thread_id()
        invocation_uuid = uuid.uuid4()
        return self._do_lock(current_thread_id, invocation_uuid)

    def try_lock(self, timeout=0):
        """Acquires the lock if it is free within the given waiting time,
        or already held by the current thread at the time of invocation and,
        the acquire limit is not exceeded, and returns the fencing token
        assigned to the current thread for this lock acquire.

        If the lock is acquired reentrantly, the same fencing token is returned.
        If the lock acquire limit is exceeded, then this method immediately returns
        :const:`INVALID_FENCE` that represents a failed lock attempt.

        If the lock is not available then the current thread becomes disabled
        for thread scheduling purposes and lies dormant until the lock is
        acquired by the current thread or the specified waiting time elapses.

        If the specified waiting time elapses, then :const:`INVALID_FENCE`
        is returned. If the time is less than or equal to zero, the method does
        not wait at all. By default, timeout is set to zero.

        A typical usage idiom for this method would be ::

            lock = client.cp_subsystem.get_lock("lock").blocking()
            fence = lock.try_lock()
            if fence != lock.INVALID_FENCE:
                try:
                    # manipulate the protected state
                finally:
                    lock.unlock()
            else:
                # perform another action

        This usage ensures that the lock is unlocked if it was acquired,
        and doesn't try to unlock if the lock was not acquired.

        See Also:
            :func:`lock` function for more information about fences.

        Args:
            timeout (int): The maximum time to wait for the lock in seconds.

        Returns:
            hazelcast.future.Future[int]: The fencing token if the lock was acquired and
            :const:`INVALID_FENCE` otherwise.

        Raises:
            LockOwnershipLostError: If the underlying CP session was
                closed before the client releases the lock
        """
        current_thread_id = thread_id()
        invocation_uuid = uuid.uuid4()
        timeout = max(0, timeout)
        return self._do_try_lock(current_thread_id, invocation_uuid, timeout)

    def unlock(self):
        """Releases the lock if the lock is currently held by the current thread.

        Returns:
            hazelcast.future.Future[None]:

        Raises:
            LockOwnershipLostError: If the underlying CP session was
                closed before the client releases the lock
            IllegalMonitorStateError: If the lock is not held by
                the current thread
        """
        current_thread_id = thread_id()
        session_id = self._get_session_id()

        # the order of the following checks is important
        try:
            self._verify_locked_session_id_if_present(current_thread_id, session_id, False)
        except LockOwnershipLostError as e:
            return ImmediateExceptionFuture(e)

        if session_id == _NO_SESSION_ID:
            self._lock_session_ids.pop(current_thread_id, None)
            return ImmediateExceptionFuture(self._new_illegal_monitor_state_error())

        def check_response(f):
            try:
                still_locked_by_the_current_thread = f.result()
                if still_locked_by_the_current_thread:
                    self._lock_session_ids[current_thread_id] = session_id
                else:
                    self._lock_session_ids.pop(current_thread_id, None)

                self._release_session(session_id)
            except SessionExpiredError:
                self._invalidate_session(session_id)
                self._lock_session_ids.pop(current_thread_id, None)
                raise self._new_lock_ownership_lost_error(session_id)
            except IllegalMonitorStateError as e:
                self._lock_session_ids.pop(current_thread_id, None)
                raise e

        return self._request_unlock(session_id, current_thread_id, uuid.uuid4()).continue_with(
            check_response
        )

    def is_locked(self):
        """Returns whether this lock is locked or not.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this lock is locked by any thread
            in the cluster, ``False`` otherwise.

        Raises:
            LockOwnershipLostError: If the underlying CP session was
                closed before the client releases the lock
        """
        current_thread_id = thread_id()
        session_id = self._get_session_id()

        try:
            self._verify_locked_session_id_if_present(current_thread_id, session_id, False)
        except LockOwnershipLostError as e:
            return ImmediateExceptionFuture(e)

        def check_response(f):
            state = _LockOwnershipState(f.result())
            if state.is_locked_by(session_id, current_thread_id):
                self._lock_session_ids[current_thread_id] = session_id
                return True

            self._verify_no_locked_session_id_present(current_thread_id)
            return state.is_locked()

        return self._request_get_lock_ownership_state().continue_with(check_response)

    def is_locked_by_current_thread(self):
        """Returns whether the lock is held by the current thread or not.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the lock is held by the current thread,
            ``False`` otherwise.

        Raises:
            LockOwnershipLostError: If the underlying CP session was
                closed before the client releases the lock
        """
        current_thread_id = thread_id()
        session_id = self._get_session_id()

        try:
            self._verify_locked_session_id_if_present(current_thread_id, session_id, False)
        except LockOwnershipLostError as e:
            return ImmediateExceptionFuture(e)

        def check_response(f):
            state = _LockOwnershipState(f.result())
            locked_by_the_current_thread = state.is_locked_by(session_id, current_thread_id)
            if locked_by_the_current_thread:
                self._lock_session_ids[current_thread_id] = session_id
            else:
                self._verify_no_locked_session_id_present(current_thread_id)

            return locked_by_the_current_thread

        return self._request_get_lock_ownership_state().continue_with(check_response)

    def get_lock_count(self):
        """Returns the reentrant lock count if the lock is held by any thread
        in the cluster.

        Returns:
            hazelcast.future.Future[int]: The reentrant lock count if the lock is held
            by any thread in the cluster

        Raises:
            LockOwnershipLostError: If the underlying CP session was
                closed before the client releases the lock
        """
        current_thread_id = thread_id()
        session_id = self._get_session_id()

        try:
            self._verify_locked_session_id_if_present(current_thread_id, session_id, False)
        except LockOwnershipLostError as e:
            return ImmediateExceptionFuture(e)

        def check_response(f):
            state = _LockOwnershipState(f.result())
            if state.is_locked_by(session_id, current_thread_id):
                self._lock_session_ids[current_thread_id] = session_id
            else:
                self._verify_no_locked_session_id_present(current_thread_id)

            return state.lock_count

        return self._request_get_lock_ownership_state().continue_with(check_response)

    def destroy(self):
        self._lock_session_ids.clear()
        return super(FencedLock, self).destroy()

    def _do_lock(self, current_thread_id, invocation_uuid):
        def do_lock_once(session_id):
            session_id = session_id.result()
            self._verify_locked_session_id_if_present(current_thread_id, session_id, True)

            def check_fence(fence):
                try:
                    fence = fence.result()
                except SessionExpiredError:
                    self._invalidate_session(session_id)
                    self._verify_no_locked_session_id_present(current_thread_id)
                    return self._do_lock(current_thread_id, invocation_uuid)
                except WaitKeyCancelledError:
                    self._release_session(session_id)
                    error = IllegalMonitorStateError(
                        "Lock(%s) not acquired because the lock call on the CP group "
                        "is cancelled, possibly because of another indeterminate call "
                        "from the same thread." % self._object_name
                    )
                    raise error
                except Exception as e:
                    self._release_session(session_id)
                    raise e

                if fence != self.INVALID_FENCE:
                    self._lock_session_ids[current_thread_id] = session_id
                    return fence

                self._release_session(session_id)
                error = LockAcquireLimitReachedError(
                    "Lock(%s) reentrant lock limit is already reached!" % self._object_name
                )
                raise error

            return self._request_lock(session_id, current_thread_id, invocation_uuid).continue_with(
                check_fence
            )

        return self._acquire_session().continue_with(do_lock_once)

    def _do_try_lock(self, current_thread_id, invocation_uuid, timeout):
        start = time.time()

        def do_try_lock_once(session_id):
            session_id = session_id.result()
            self._verify_locked_session_id_if_present(current_thread_id, session_id, True)

            def check_fence(fence):
                try:
                    fence = fence.result()
                except SessionExpiredError:
                    self._invalidate_session(session_id)
                    self._verify_no_locked_session_id_present(current_thread_id)

                    remaining_timeout = timeout - (time.time() - start)
                    if remaining_timeout <= 0:
                        return self.INVALID_FENCE
                    return self._do_try_lock(current_thread_id, invocation_uuid, remaining_timeout)
                except WaitKeyCancelledError:
                    self._release_session(session_id)
                    return self.INVALID_FENCE
                except Exception as e:
                    self._release_session(session_id)
                    raise e

                if fence != self.INVALID_FENCE:
                    self._lock_session_ids[current_thread_id] = session_id
                else:
                    self._release_session(session_id)

                return fence

            return self._request_try_lock(
                session_id, current_thread_id, invocation_uuid, timeout
            ).continue_with(check_fence)

        return self._acquire_session().continue_with(do_try_lock_once)

    def _verify_locked_session_id_if_present(self, current_thread_id, session_id, release_session):
        lock_session_id = self._lock_session_ids.get(current_thread_id, None)
        if lock_session_id and lock_session_id != session_id:
            self._lock_session_ids.pop(current_thread_id, None)
            if release_session:
                self._release_session(session_id)

            raise self._new_lock_ownership_lost_error(lock_session_id)

    def _verify_no_locked_session_id_present(self, current_thread_id):
        lock_session_id = self._lock_session_ids.pop(current_thread_id, None)
        if lock_session_id:
            raise self._new_lock_ownership_lost_error(lock_session_id)

    def _new_lock_ownership_lost_error(self, lock_session_id):
        error = LockOwnershipLostError(
            "Current thread is not the owner of the Lock(%s) because its "
            "Session(%s) is closed by the server." % (self._proxy_name, lock_session_id)
        )
        return error

    def _new_illegal_monitor_state_error(self):
        error = IllegalMonitorStateError(
            "Current thread is not the owner of the Lock(%s)" % self._proxy_name
        )
        return error

    def _request_lock(self, session_id, current_thread_id, invocation_uuid):
        codec = fenced_lock_lock_codec
        request = codec.encode_request(
            self._group_id, self._object_name, session_id, current_thread_id, invocation_uuid
        )
        return self._invoke(request, codec.decode_response)

    def _request_try_lock(self, session_id, current_thread_id, invocation_uuid, timeout):
        codec = fenced_lock_try_lock_codec
        request = codec.encode_request(
            self._group_id,
            self._object_name,
            session_id,
            current_thread_id,
            invocation_uuid,
            to_millis(timeout),
        )
        return self._invoke(request, codec.decode_response)

    def _request_unlock(self, session_id, current_thread_id, invocation_uuid):
        codec = fenced_lock_unlock_codec
        request = codec.encode_request(
            self._group_id, self._object_name, session_id, current_thread_id, invocation_uuid
        )
        return self._invoke(request, codec.decode_response)

    def _request_get_lock_ownership_state(self):
        codec = fenced_lock_get_lock_ownership_codec
        request = codec.encode_request(self._group_id, self._object_name)
        return self._invoke(request, codec.decode_response)


class _LockOwnershipState(object):
    def __init__(self, state):
        self.fence = state["fence"]
        self.lock_count = state["lock_count"]
        self.session_id = state["session_id"]
        self.thread_id = state["thread_id"]

    def is_locked(self):
        return self.fence != FencedLock.INVALID_FENCE

    def is_locked_by(self, session_id, current_thread_id):
        return (
            self.is_locked()
            and self.session_id == session_id
            and self.thread_id == current_thread_id
        )



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer, RESPONSE_HEADER_SIZE

# hex: 0x150200
_REQUEST_MESSAGE_TYPE = 1376768
# hex: 0x150201
_RESPONSE_MESSAGE_TYPE = 1376769

_REQUEST_TIMEOUT_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_DURABILITY_OFFSET = _REQUEST_TIMEOUT_OFFSET + LONG_SIZE_IN_BYTES
_REQUEST_TRANSACTION_TYPE_OFFSET = _REQUEST_DURABILITY_OFFSET + INT_SIZE_IN_BYTES
_REQUEST_THREAD_ID_OFFSET = _REQUEST_TRANSACTION_TYPE_OFFSET + INT_SIZE_IN_BYTES
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_THREAD_ID_OFFSET + LONG_SIZE_IN_BYTES
_RESPONSE_RESPONSE_OFFSET = RESPONSE_HEADER_SIZE


def encode_request(timeout, durability, transaction_type, thread_id):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE, True)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_TIMEOUT_OFFSET, timeout)
    FixSizedTypesCodec.encode_int(buf, _REQUEST_DURABILITY_OFFSET, durability)
    FixSizedTypesCodec.encode_int(buf, _REQUEST_TRANSACTION_TYPE_OFFSET, transaction_type)
    FixSizedTypesCodec.encode_long(buf, _REQUEST_THREAD_ID_OFFSET, thread_id)
    return OutboundMessage(buf, False)


def decode_response(msg):
    initial_frame = msg.next_frame()
    return FixSizedTypesCodec.decode_uuid(initial_frame.buf, _RESPONSE_RESPONSE_OFFSET)



from hazelcast.serialization.bits import *
from hazelcast.protocol.builtin import FixSizedTypesCodec
from hazelcast.protocol.client_message import OutboundMessage, REQUEST_HEADER_SIZE, create_initial_buffer, RESPONSE_HEADER_SIZE, EVENT_HEADER_SIZE
from hazelcast.protocol.builtin import StringCodec
from hazelcast.protocol.builtin import DataCodec
from hazelcast.protocol.builtin import CodecUtil
from hazelcast.protocol.builtin import ListMultiFrameCodec
from hazelcast.protocol.builtin import ListUUIDCodec
from hazelcast.protocol.builtin import ListLongCodec

# hex: 0x013F00
_REQUEST_MESSAGE_TYPE = 81664
# hex: 0x013F01
_RESPONSE_MESSAGE_TYPE = 81665
# hex: 0x013F02
_EVENT_I_MAP_INVALIDATION_MESSAGE_TYPE = 81666
# hex: 0x013F03
_EVENT_I_MAP_BATCH_INVALIDATION_MESSAGE_TYPE = 81667

_REQUEST_LISTENER_FLAGS_OFFSET = REQUEST_HEADER_SIZE
_REQUEST_LOCAL_ONLY_OFFSET = _REQUEST_LISTENER_FLAGS_OFFSET + INT_SIZE_IN_BYTES
_REQUEST_INITIAL_FRAME_SIZE = _REQUEST_LOCAL_ONLY_OFFSET + BOOLEAN_SIZE_IN_BYTES
_RESPONSE_RESPONSE_OFFSET = RESPONSE_HEADER_SIZE
_EVENT_I_MAP_INVALIDATION_SOURCE_UUID_OFFSET = EVENT_HEADER_SIZE
_EVENT_I_MAP_INVALIDATION_PARTITION_UUID_OFFSET = _EVENT_I_MAP_INVALIDATION_SOURCE_UUID_OFFSET + UUID_SIZE_IN_BYTES
_EVENT_I_MAP_INVALIDATION_SEQUENCE_OFFSET = _EVENT_I_MAP_INVALIDATION_PARTITION_UUID_OFFSET + UUID_SIZE_IN_BYTES


def encode_request(name, listener_flags, local_only):
    buf = create_initial_buffer(_REQUEST_INITIAL_FRAME_SIZE, _REQUEST_MESSAGE_TYPE)
    FixSizedTypesCodec.encode_int(buf, _REQUEST_LISTENER_FLAGS_OFFSET, listener_flags)
    FixSizedTypesCodec.encode_boolean(buf, _REQUEST_LOCAL_ONLY_OFFSET, local_only)
    StringCodec.encode(buf, name, True)
    return OutboundMessage(buf, False)


def decode_response(msg):
    initial_frame = msg.next_frame()
    return FixSizedTypesCodec.decode_uuid(initial_frame.buf, _RESPONSE_RESPONSE_OFFSET)


def handle(msg, handle_i_map_invalidation_event=None, handle_i_map_batch_invalidation_event=None):
    message_type = msg.get_message_type()
    if message_type == _EVENT_I_MAP_INVALIDATION_MESSAGE_TYPE and handle_i_map_invalidation_event is not None:
        initial_frame = msg.next_frame()
        source_uuid = FixSizedTypesCodec.decode_uuid(initial_frame.buf, _EVENT_I_MAP_INVALIDATION_SOURCE_UUID_OFFSET)
        partition_uuid = FixSizedTypesCodec.decode_uuid(initial_frame.buf, _EVENT_I_MAP_INVALIDATION_PARTITION_UUID_OFFSET)
        sequence = FixSizedTypesCodec.decode_long(initial_frame.buf, _EVENT_I_MAP_INVALIDATION_SEQUENCE_OFFSET)
        key = CodecUtil.decode_nullable(msg, DataCodec.decode)
        handle_i_map_invalidation_event(key, source_uuid, partition_uuid, sequence)
        return
    if message_type == _EVENT_I_MAP_BATCH_INVALIDATION_MESSAGE_TYPE and handle_i_map_batch_invalidation_event is not None:
        msg.next_frame()
        keys = ListMultiFrameCodec.decode(msg, DataCodec.decode)
        source_uuids = ListUUIDCodec.decode(msg)
        partition_uuids = ListUUIDCodec.decode(msg)
        sequences = ListLongCodec.decode(msg)
        handle_i_map_batch_invalidation_event(keys, source_uuids, partition_uuids, sequences)
        return



import sys
import time
from threading import RLock, Lock

from hazelcast.errors import (
    SessionExpiredError,
    CPGroupDestroyedError,
    HazelcastClientNotActiveError,
)
from hazelcast.future import ImmediateExceptionFuture, ImmediateFuture, combine_futures
from hazelcast.invocation import Invocation
from hazelcast.protocol.codec import (
    cp_group_create_cp_group_codec,
    cp_session_heartbeat_session_codec,
    cp_session_create_session_codec,
    cp_session_close_session_codec,
    cp_session_generate_thread_id_codec,
    semaphore_get_semaphore_type_codec,
)
from hazelcast.proxy.cp.atomic_long import AtomicLong
from hazelcast.proxy.cp.atomic_reference import AtomicReference
from hazelcast.proxy.cp.count_down_latch import CountDownLatch
from hazelcast.proxy.cp.fenced_lock import FencedLock
from hazelcast.proxy.cp.semaphore import SessionAwareSemaphore, SessionlessSemaphore
from hazelcast.util import check_true, AtomicInteger, thread_id


class CPSubsystem(object):
    """CP Subsystem is a component of Hazelcast that builds a strongly consistent
    layer for a set of distributed data structures.

    Its APIs can be used for implementing distributed coordination use cases,
    such as leader election, distributed locking, synchronization, and metadata
    management.

    Its data structures are CP with respect to the CAP principle, i.e., they
    always maintain linearizability and prefer consistency over availability
    during network partitions. Besides network partitions, CP Subsystem
    withstands server and client failures.

    Data structures in CP Subsystem run in CP groups. Each CP group elects
    its own Raft leader and runs the Raft consensus algorithm independently.

    The CP data structures differ from the other Hazelcast data structures
    in two aspects. First, an internal commit is performed on the METADATA CP
    group every time you fetch a proxy from this interface. Hence, callers
    should cache returned proxy objects. Second, if you call ``destroy()``
    on a CP data structure proxy, that data structure is terminated on the
    underlying CP group and cannot be reinitialized until the CP group is
    force-destroyed. For this reason, please make sure that you are completely
    done with a CP data structure before destroying its proxy.
    """

    def __init__(self, context):
        self._proxy_manager = CPProxyManager(context)

    def get_atomic_long(self, name):
        """Returns the distributed AtomicLong instance with given name.

        The instance is created on CP Subsystem.

        If no group name is given within the ``name`` argument, then the
        AtomicLong instance will be created on the default CP group.
        If a group name is given, like ``.get_atomic_long("myLong@group1")``,
        the given group will be initialized first, if not initialized
        already, and then the instance will be created on this group.

        Args:
            name (str): Name of the AtomicLong.

        Returns:
            hazelcast.proxy.cp.atomic_long.AtomicLong: The AtomicLong proxy
            for the given name.
        """
        return self._proxy_manager.get_or_create(ATOMIC_LONG_SERVICE, name)

    def get_atomic_reference(self, name):
        """Returns the distributed AtomicReference instance with given name.

        The instance is created on CP Subsystem.

        If no group name is given within the ``name`` argument, then the
        AtomicLong instance will be created on the DEFAULT CP group.
        If a group name is given, like ``.get_atomic_reference("myRef@group1")``,
        the given group will be initialized first, if not initialized
        already, and then the instance will be created on this group.

        Args:
            name (str): Name of the AtomicReference.

        Returns:
            hazelcast.proxy.cp.atomic_reference.AtomicReference: The AtomicReference
            proxy for the given name.
        """
        return self._proxy_manager.get_or_create(ATOMIC_REFERENCE_SERVICE, name)

    def get_count_down_latch(self, name):
        """Returns the distributed CountDownLatch instance with given name.

        The instance is created on CP Subsystem.

        If no group name is given within the ``name`` argument, then the
        CountDownLatch instance will be created on the DEFAULT CP group.
        If a group name is given, like ``.get_count_down_latch("myLatch@group1")``,
        the given group will be initialized first, if not initialized
        already, and then the instance will be created on this group.

        Args:
            name (str): Name of the CountDownLatch.

        Returns:
            hazelcast.proxy.cp.count_down_latch.CountDownLatch: The CountDownLatch
            proxy for the given name.
        """
        return self._proxy_manager.get_or_create(COUNT_DOWN_LATCH_SERVICE, name)

    def get_lock(self, name):
        """Returns the distributed FencedLock instance instance with given name.

        The instance is created on CP Subsystem.

        If no group name is given within the ``name`` argument, then the
        FencedLock instance will be created on the DEFAULT CP group.
        If a group name is given, like ``.get_lock("myLock@group1")``,
        the given group will be initialized first, if not initialized
        already, and then the instance will be created on this group.

        Args:
            name (str): Name of the FencedLock

        Returns:
            hazelcast.proxy.cp.fenced_lock.FencedLock: The FencedLock proxy
            for the given name.
        """
        return self._proxy_manager.get_or_create(LOCK_SERVICE, name)

    def get_semaphore(self, name):
        """Returns the distributed Semaphore instance instance with given name.

        The instance is created on CP Subsystem.

        If no group name is given within the ``name`` argument, then the
        Semaphore instance will be created on the DEFAULT CP group.
        If a group name is given, like ``.get_semaphore("mySemaphore@group1")``,
        the given group will be initialized first, if not initialized
        already, and then the instance will be created on this group.

        Args:
            name (str): Name of the Semaphore

        Returns:
            hazelcast.proxy.cp.semaphore.Semaphore: The Semaphore proxy for the given name.
        """
        return self._proxy_manager.get_or_create(SEMAPHORE_SERVICE, name)


_DEFAULT_GROUP_NAME = "default"
_METADATA_CP_GROUP_NAME = "metadata"


def _without_default_group_name(name):
    name = name.strip()
    idx = name.find("@")
    if idx == -1:
        return name

    check_true(name.find("@", idx + 1) == -1, "Custom group name must be specified at most once")
    group_name = name[idx + 1 :].strip()
    check_true(
        group_name.lower() != _METADATA_CP_GROUP_NAME,
        "CP data structures cannot run on the METADATA CP group!",
    )
    if group_name.lower() == _DEFAULT_GROUP_NAME:
        return name[:idx]
    return name


def _get_object_name_for_proxy(name):
    idx = name.find("@")
    if idx == -1:
        return name

    group_name = name[idx + 1 :].strip()
    check_true(len(group_name) > 0, "Custom CP group name cannot be empty string")
    object_name = name[:idx].strip()
    check_true(len(object_name) > 0, "Object name cannot be empty string")
    return object_name


ATOMIC_LONG_SERVICE = "hz:raft:atomicLongService"
ATOMIC_REFERENCE_SERVICE = "hz:raft:atomicRefService"
COUNT_DOWN_LATCH_SERVICE = "hz:raft:countDownLatchService"
LOCK_SERVICE = "hz:raft:lockService"
SEMAPHORE_SERVICE = "hz:raft:semaphoreService"


class CPProxyManager(object):
    def __init__(self, context):
        self._context = context
        self._lock_proxies = dict()  # proxy_name to FencedLock
        self._mux = Lock()  # Guards the _lock_proxies

    def get_or_create(self, service_name, proxy_name):
        proxy_name = _without_default_group_name(proxy_name)
        object_name = _get_object_name_for_proxy(proxy_name)

        group_id = self._get_group_id(proxy_name)
        if service_name == ATOMIC_LONG_SERVICE:
            return AtomicLong(self._context, group_id, service_name, proxy_name, object_name)
        elif service_name == ATOMIC_REFERENCE_SERVICE:
            return AtomicReference(self._context, group_id, service_name, proxy_name, object_name)
        elif service_name == COUNT_DOWN_LATCH_SERVICE:
            return CountDownLatch(self._context, group_id, service_name, proxy_name, object_name)
        elif service_name == LOCK_SERVICE:
            return self._create_fenced_lock(group_id, proxy_name, object_name)
        elif service_name == SEMAPHORE_SERVICE:
            return self._create_semaphore(group_id, proxy_name, object_name)
        else:
            raise ValueError("Unknown service name: %s" % service_name)

    def _create_fenced_lock(self, group_id, proxy_name, object_name):
        with self._mux:
            proxy = self._lock_proxies.get(proxy_name, None)
            if proxy:
                if proxy.get_group_id() != group_id:
                    self._lock_proxies.pop(proxy_name, None)
                else:
                    return proxy

            proxy = FencedLock(self._context, group_id, LOCK_SERVICE, proxy_name, object_name)
            self._lock_proxies[proxy_name] = proxy
            return proxy

    def _create_semaphore(self, group_id, proxy_name, object_name):
        codec = semaphore_get_semaphore_type_codec
        request = codec.encode_request(proxy_name)
        invocation = Invocation(request, response_handler=codec.decode_response)
        invocation_service = self._context.invocation_service
        invocation_service.invoke(invocation)
        jdk_compatible = invocation.future.result()
        if jdk_compatible:
            return SessionlessSemaphore(
                self._context, group_id, SEMAPHORE_SERVICE, proxy_name, object_name
            )
        else:
            return SessionAwareSemaphore(
                self._context, group_id, SEMAPHORE_SERVICE, proxy_name, object_name
            )

    def _get_group_id(self, proxy_name):
        codec = cp_group_create_cp_group_codec
        request = codec.encode_request(proxy_name)
        invocation = Invocation(request, response_handler=codec.decode_response)
        invocation_service = self._context.invocation_service
        invocation_service.invoke(invocation)
        return invocation.future.result()


class _SessionState(object):
    __slots__ = ("id", "group_id", "ttl", "creation_time", "acquire_count")

    def __init__(self, state_id, group_id, ttl):
        self.id = state_id
        self.ttl = ttl
        self.group_id = group_id
        self.creation_time = time.time()
        self.acquire_count = AtomicInteger()

    def acquire(self, count):
        self.acquire_count.add(count)
        return self.id

    def release(self, count):
        self.acquire_count.add(-count)

    def is_valid(self):
        return self.is_in_use() or not self._is_expired(time.time())

    def is_in_use(self):
        return self.acquire_count.get() > 0

    def _is_expired(self, timestamp):
        expiration_time = self.creation_time + self.ttl
        if expiration_time < 0:
            expiration_time = sys.maxsize
        return timestamp > expiration_time

    def __eq__(self, other):
        return isinstance(other, _SessionState) and self.id == other.id

    def __ne__(self, other):
        return not self.__eq__(other)


_NO_SESSION_ID = -1


class ProxySessionManager(object):
    def __init__(self, context):
        self._context = context
        self._mutexes = dict()  # RaftGroupId to RLock
        self._sessions = dict()  # RaftGroupId to SessionState
        self._thread_ids = dict()  # (RaftGroupId, thread_id) to global thread id
        self._heartbeat_timer = None
        self._shutdown = False
        self._lock = RLock()

    def get_session_id(self, group_id):
        session = self._sessions.get(group_id, None)
        if session:
            return session.id
        return _NO_SESSION_ID

    def acquire_session(self, group_id, count):
        return self._get_or_create_session(group_id).continue_with(
            lambda state: state.result().acquire(count)
        )

    def release_session(self, group_id, session_id, count):
        session = self._sessions.get(group_id, None)
        if session and session.id == session_id:
            session.release(count)

    def invalidate_session(self, group_id, session_id):
        # called from the reactor thread only
        session = self._sessions.get(group_id, None)
        if session and session.id == session_id:
            self._sessions.pop(group_id, None)

    def get_or_create_unique_thread_id(self, group_id):
        with self._lock:
            if self._shutdown:
                error = HazelcastClientNotActiveError("Session manager is already shut down!")
                return ImmediateExceptionFuture(error)

            key = (group_id, thread_id())
            global_thread_id = self._thread_ids.get(key)
            if global_thread_id:
                return ImmediateFuture(global_thread_id)

            return self._request_generate_thread_id(group_id).continue_with(
                lambda t_id: self._thread_ids.setdefault(key, t_id.result())
            )

    def shutdown(self):
        with self._lock:
            if self._shutdown:
                return ImmediateFuture(None)

            self._shutdown = True
            if self._heartbeat_timer:
                self._heartbeat_timer.cancel()

            futures = []
            for session in list(self._sessions.values()):
                future = self._request_close_session(session.group_id, session.id)
                futures.append(future)

            def clear(_):
                self._sessions.clear()
                self._mutexes.clear()
                self._thread_ids.clear()

            return combine_futures(futures).continue_with(clear)

    def _request_generate_thread_id(self, group_id):
        codec = cp_session_generate_thread_id_codec
        request = codec.encode_request(group_id)
        invocation = Invocation(request, response_handler=codec.decode_response)
        self._context.invocation_service.invoke(invocation)
        return invocation.future

    def _request_close_session(self, group_id, session_id):
        codec = cp_session_close_session_codec
        request = codec.encode_request(group_id, session_id)
        invocation = Invocation(request, response_handler=codec.decode_response)
        self._context.invocation_service.invoke(invocation)
        return invocation.future

    def _get_or_create_session(self, group_id):
        with self._lock:
            if self._shutdown:
                error = HazelcastClientNotActiveError("Session manager is already shut down!")
                return ImmediateExceptionFuture(error)

            session = self._sessions.get(group_id, None)
            if session is None or not session.is_valid():
                with self._mutex(group_id):
                    session = self._sessions.get(group_id)
                    if session is None or not session.is_valid():
                        return self._create_new_session(group_id)
            return ImmediateFuture(session)

    def _create_new_session(self, group_id):
        f = self._request_new_session(group_id)
        return f.continue_with(self._do_create_new_session, group_id)

    def _do_create_new_session(self, response, group_id):
        # called from the reactor thread only
        response = response.result()
        session = _SessionState(response["session_id"], group_id, response["ttl_millis"] / 1000.0)
        self._sessions[group_id] = session
        self._start_heartbeat_timer(response["heartbeat_millis"] / 1000.0)
        return session

    def _request_new_session(self, group_id):
        codec = cp_session_create_session_codec
        request = codec.encode_request(group_id, self._context.name)
        invocation = Invocation(request, response_handler=codec.decode_response)
        self._context.invocation_service.invoke(invocation)
        return invocation.future

    def _mutex(self, group_id):
        mutex = self._mutexes.get(group_id, None)
        if mutex:
            return mutex

        mutex = RLock()
        current = self._mutexes.setdefault(group_id, mutex)
        return current

    def _start_heartbeat_timer(self, period):
        if self._heartbeat_timer is not None:
            return

        def heartbeat():
            if self._shutdown:
                return

            for session in list(self._sessions.values()):
                if session.is_in_use():

                    def cb(heartbeat_future, session=session):
                        if heartbeat_future.is_success():
                            return

                        error = heartbeat_future.exception()
                        if isinstance(error, (SessionExpiredError, CPGroupDestroyedError)):
                            self.invalidate_session(session.group_id, session.id)

                    f = self._request_heartbeat(session.group_id, session.id)
                    f.add_done_callback(cb)

            r = self._context.reactor
            self._heartbeat_timer = r.add_timer(period, heartbeat)

        reactor = self._context.reactor
        self._heartbeat_timer = reactor.add_timer(period, heartbeat)

    def _request_heartbeat(self, group_id, session_id):
        codec = cp_session_heartbeat_session_codec
        request = codec.encode_request(group_id, session_id)
        invocation = Invocation(request)
        self._context.invocation_service.invoke(invocation)
        return invocation.future



from hazelcast.protocol.codec import (
    set_add_all_codec,
    set_add_codec,
    set_add_listener_codec,
    set_clear_codec,
    set_compare_and_remove_all_codec,
    set_compare_and_retain_all_codec,
    set_contains_all_codec,
    set_contains_codec,
    set_get_all_codec,
    set_is_empty_codec,
    set_remove_codec,
    set_remove_listener_codec,
    set_size_codec,
)

from hazelcast.proxy.base import PartitionSpecificProxy, ItemEvent, ItemEventType
from hazelcast.util import check_not_none, ImmutableLazyDataList


class Set(PartitionSpecificProxy):
    """Concurrent, distributed implementation of Set"""

    def add(self, item):
        """Adds the specified item if it is not exists in this set.

        Args:
            item: The specified item to be added.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this set is changed after call, ``False`` otherwise.
        """
        check_not_none(item, "Value can't be None")
        element_data = self._to_data(item)
        request = set_add_codec.encode_request(self.name, element_data)
        return self._invoke(request, set_add_codec.decode_response)

    def add_all(self, items):
        """Adds the elements in the specified collection if they're not exist in this set.

        Args:
            items (list): Collection which includes the items to be added.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this set is changed after call, ``False`` otherwise.
        """
        check_not_none(items, "Value can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "Value can't be None")
            data_items.append(self._to_data(item))

        request = set_add_all_codec.encode_request(self.name, data_items)
        return self._invoke(request, set_add_all_codec.decode_response)

    def add_listener(self, include_value=False, item_added_func=None, item_removed_func=None):
        """Adds an item listener for this container.

        Listener will be notified for all container add/remove events.

        Args:
            include_value (bool): Whether received events include the updated item or not.
            item_added_func (function): Function to be called when an item is added to this set.
            item_removed_func (function): Function to be called when an item is deleted from this set.

        Returns:
            hazelcast.future.Future[str]: A registration id which is used as a key to remove the listener.
        """
        request = set_add_listener_codec.encode_request(self.name, include_value, self._is_smart)

        def handle_event_item(item, uuid, event_type):
            item = item if include_value else None
            member = self._context.cluster_service.get_member(uuid)

            item_event = ItemEvent(self.name, item, event_type, member, self._to_object)
            if event_type == ItemEventType.ADDED:
                if item_added_func:
                    item_added_func(item_event)
            else:
                if item_removed_func:
                    item_removed_func(item_event)

        return self._register_listener(
            request,
            lambda r: set_add_listener_codec.decode_response(r),
            lambda reg_id: set_remove_listener_codec.encode_request(self.name, reg_id),
            lambda m: set_add_listener_codec.handle(m, handle_event_item),
        )

    def clear(self):
        """Clears the set. Set will be empty with this call.

        Returns:
            hazelcast.future.Future[None]:
        """
        request = set_clear_codec.encode_request(self.name)
        return self._invoke(request)

    def contains(self, item):
        """Determines whether this set contains the specified item or not.

        Args:
            item: The specified item to be searched.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the specified item exists in this set, ``False`` otherwise.
        """
        check_not_none(item, "Value can't be None")
        item_data = self._to_data(item)
        request = set_contains_codec.encode_request(self.name, item_data)
        return self._invoke(request, set_contains_codec.decode_response)

    def contains_all(self, items):
        """Determines whether this set contains all of the items in the specified collection or not.

        Args:
            items (list): The specified collection which includes the items to be searched.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if all of the items in the specified collection exist in this set,
            ``False`` otherwise.
        """
        check_not_none(items, "Value can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "Value can't be None")
            data_items.append(self._to_data(item))

        request = set_contains_all_codec.encode_request(self.name, data_items)
        return self._invoke(request, set_contains_all_codec.decode_response)

    def get_all(self):
        """Returns all of the items in the set.

        Returns:
            hazelcast.future.Future[list]: List of the items in this set.
        """

        def handler(message):
            return ImmutableLazyDataList(
                set_get_all_codec.decode_response(message), self._to_object
            )

        request = set_get_all_codec.encode_request(self.name)
        return self._invoke(request, handler)

    def is_empty(self):
        """Determines whether this set is empty or not.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this set is empty, ``False`` otherwise.
        """
        request = set_is_empty_codec.encode_request(self.name)
        return self._invoke(request, set_is_empty_codec.decode_response)

    def remove(self, item):
        """Removes the specified element from the set if it exists.

        Args:
            item: The specified element to be removed.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the specified element exists in this set, ``False`` otherwise.
        """
        check_not_none(item, "Value can't be None")
        item_data = self._to_data(item)
        request = set_remove_codec.encode_request(self.name, item_data)
        return self._invoke(request, set_remove_codec.decode_response)

    def remove_all(self, items):
        """Removes all of the elements of the specified collection from this set.

        Args:
            items (list): The specified collection.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the call changed this set, ``False`` otherwise.
        """
        check_not_none(items, "Value can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "Value can't be None")
            data_items.append(self._to_data(item))

        request = set_compare_and_remove_all_codec.encode_request(self.name, data_items)
        return self._invoke(request, set_compare_and_remove_all_codec.decode_response)

    def remove_listener(self, registration_id):
        """Removes the specified item listener.

        Returns silently if the specified listener was not added before.

        Args:
            registration_id (str): Id of the listener to be deleted.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if the item listener is removed, ``False`` otherwise.
        """
        return self._deregister_listener(registration_id)

    def retain_all(self, items):
        """Removes the items which are not contained in the specified collection.

        In other words, only the items that are contained in the specified collection will be retained.

        Args:
            items (list): Collection which includes the elements to be retained in this set.

        Returns:
            hazelcast.future.Future[bool]: ``True`` if this set changed as a result of the call, ``False`` otherwise.
        """
        check_not_none(items, "Value can't be None")
        data_items = []
        for item in items:
            check_not_none(item, "Value can't be None")
            data_items.append(self._to_data(item))

        request = set_compare_and_retain_all_codec.encode_request(self.name, data_items)
        return self._invoke(request, set_compare_and_retain_all_codec.decode_response)

    def size(self):
        """Returns the number of items in this set.

        Returns:
            hazelcast.future.Future[int]: Number of items in this set.
        """
        request = set_size_codec.encode_request(self.name)
        return self._invoke(request, set_size_codec.decode_response)
